"""Regression tests for plotting functions, by comparing to saved baseline images."""

import warnings

import matplotlib.pyplot as plt
import numpy as np
import pytest

from desc.basis import (
    DoubleFourierSeries,
    FourierSeries,
    FourierZernikeBasis,
    PowerSeries,
)
from desc.coils import CoilSet, FourierXYZCoil, MixedCoilSet
from desc.compute import data_index
from desc.examples import get
from desc.geometry import FourierRZToroidalSurface, FourierXYZCurve
from desc.grid import ConcentricGrid, Grid, LinearGrid, QuadratureGrid
from desc.integrals import surface_averages
from desc.io import load
from desc.magnetic_fields import OmnigenousField, ToroidalMagneticField
from desc.plotting import (
    plot_1d,
    plot_2d,
    plot_3d,
    plot_basis,
    plot_boozer_modes,
    plot_boozer_surface,
    plot_boundaries,
    plot_boundary,
    plot_coefficients,
    plot_coils,
    plot_comparison,
    plot_fsa,
    plot_grid,
    plot_logo,
    plot_qs_error,
    plot_section,
    plot_surfaces,
    poincare_plot,
)
from desc.utils import isalmostequal

tol_1d = 4.5
tol_2d = 10
tol_3d = 10


@pytest.mark.unit
def test_kwarg_warning(DummyStellarator):
    """Test that passing in unknown kwargs throws an error."""
    eq = load(load_from=str(DummyStellarator["output_path"]))
    with pytest.raises(AssertionError):
        fig, ax = plot_1d(eq, "psi_rr", not_a_kwarg=True)
    return None


@pytest.mark.unit
def test_kwarg_future_warning(DummyStellarator):
    """Test that passing in deprecated kwargs throws a warning."""
    eq = load(load_from=str(DummyStellarator["output_path"]))
    with pytest.raises(FutureWarning):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            fig, ax = plot_surfaces(eq, zeta=2)
    return None


class TestPlot1D:
    """Tests for plot_1d."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_1d_p(self):
        """Test plotting 1d pressure profile."""
        eq = get("SOLOVEV")
        fig, ax, data = plot_1d(eq, "p", figsize=(4, 4), return_data=True)
        assert "p" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_1d_elongation(self):
        """Test plotting 1d elongation as a function of toroidal angle."""
        eq = get("precise_QA")
        grid = LinearGrid(N=20, NFP=eq.NFP)
        fig, ax, data = plot_1d(
            eq, "a_major/a_minor", grid=grid, figsize=(4, 4), return_data=True
        )
        assert "a_major/a_minor" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_1d_iota(self):
        """Test plotting 1d rotational transform."""
        eq = get("DSHAPE_current")
        grid = LinearGrid(rho=0.5, theta=100, zeta=0.0)
        fig, ax, data = plot_1d(eq, "iota", grid=grid, figsize=(4, 4), return_data=True)
        assert "theta" in data.keys()
        assert "iota" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_1d_iota_radial(self):
        """Test plotting 1d rotational transform."""
        eq = get("DSHAPE_current")
        fig, ax, data = plot_1d(eq, "iota", figsize=(4, 4), return_data=True)
        assert "rho" in data.keys()
        assert "iota" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_1d_logpsi(self):
        """Test plotting 1d flux function with log scale."""
        eq = get("DSHAPE_current")
        fig, ax, data = plot_1d(eq, "psi", log=True, figsize=(4, 4), return_data=True)
        ax.set_ylim([1e-5, 1e0])
        assert "rho" in data.keys()
        assert "psi" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_1d_curve(self):
        """Test plot_1d function for Curve objects."""
        curve = FourierXYZCurve([0, 10, 1])
        fig, ax = plot_1d(curve, "curvature")
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_1d_surface(self):
        """Test plot_1d function for Surface objects."""
        surf = FourierRZToroidalSurface()
        fig, ax = plot_1d(surf, "curvature_H_rho", grid=LinearGrid(M=50))
        return fig


class TestPlot2D:
    """Tests for plot_2d."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=10)
    def test_2d_logF(self):
        """Test plotting 2d force error with log scale."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=100, theta=100, zeta=0.0)
        fig, ax, data = plot_2d(
            eq, "|F|", log=True, grid=grid, figsize=(4, 4), return_data=True
        )
        assert "|F|" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_2d_g_tz(self):
        """Test plotting 2d metric coefficients vs theta/zeta."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=0.5, theta=100, zeta=100)
        fig, ax, data = plot_2d(
            eq, "sqrt(g)", grid=grid, figsize=(4, 4), return_data=True
        )
        assert "theta" in data.keys()
        assert "zeta" in data.keys()

        assert "sqrt(g)" in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_2d_g_rz(self):
        """Test plotting 2d metric coefficients vs rho/zeta."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=100, theta=0.0, zeta=100)
        fig, ax, data = plot_2d(
            eq, "sqrt(g)", grid=grid, figsize=(4, 4), return_data=True
        )
        assert "rho" in data.keys()
        assert "zeta" in data.keys()
        assert "sqrt(g)" in data.keys()

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_con_basis(self):
        """Test 2d plot of R component of e^rho."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_2d(
            eq, "e^rho", component="R", figsize=(4, 4), return_data=True
        )
        for string in ["e^rho", "normalization", "theta", "zeta"]:
            assert string in data.keys()
        assert data["normalization"] == 1

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_cov_basis(self):
        """Test 2d plot of norm of e_rho."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_2d(eq, "e_rho", figsize=(4, 4), return_data=True)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_normF_2d(self):
        """Test 2d plot of normalized force."""
        grid = LinearGrid(rho=np.array(0.8), M=20, N=2)
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_2d(
            eq, "|F|", norm_F=True, figsize=(4, 4), return_data=True, grid=grid
        )
        for string in ["|F|", "theta", "zeta"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_2d_surface(self):
        """Test plot_2d function for Surface objects."""
        surf = FourierRZToroidalSurface()
        fig, ax = plot_2d(surf, "curvature_H_rho")
        return fig

    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    @pytest.mark.unit
    def test_2d_plot_Bn(self):
        """Test 2d plotting of Bn on equilibrium surface."""
        eq = get("HELIOTRON")
        fig, _ = plot_2d(
            eq,
            "B*n",
            field=ToroidalMagneticField(1, 1),
            field_grid=LinearGrid(M=10, N=10),
            grid=LinearGrid(M=30, N=30, NFP=eq.NFP, endpoint=True),
        )
        return fig


class TestPlot3D:
    """Tests for plot_3d."""

    @pytest.mark.unit
    def test_3d_tz(self):
        """Test 3d plot of force on interior surface."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=0.5, theta=100, zeta=100)
        fig, data = plot_3d(eq, "|F|", log=True, grid=grid, return_data=True)
        assert "X" in data.keys()
        assert "Y" in data.keys()
        assert "Z" in data.keys()

        assert "|F|" in data.keys()
        return fig

    @pytest.mark.unit
    def test_3d_rz(self):
        """Test 3d plotting of pressure on toroidal cross section."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=100, theta=0.0, zeta=100)
        fig = plot_3d(eq, "p", grid=grid)
        return fig

    @pytest.mark.unit
    def test_3d_rt(self):
        """Test 3d plotting of flux on poloidal ribbon."""
        eq = get("DSHAPE_CURRENT")
        grid = LinearGrid(rho=100, theta=100, zeta=0.0)
        fig = plot_3d(eq, "psi", grid=grid)
        return fig

    @pytest.mark.unit
    def test_plot_3d_surface(self):
        """Test 3d plotting of surface object."""
        surf = FourierRZToroidalSurface()
        fig = plot_3d(
            surf,
            "curvature_H_rho",
            showgrid=False,
            showscale=False,
            zeroline=False,
            showticklabels=False,
            showaxislabels=False,
        )
        return fig

    @pytest.mark.unit
    def test_3d_plot_Bn(self):
        """Test 3d plotting of Bn on equilibrium surface."""
        eq = get("precise_QA")
        with pytest.warns(UserWarning, match="Reducing radial"):
            eq.change_resolution(M=4, N=4, L=4, M_grid=8, N_grid=8, L_grid=8)
        fig = plot_3d(
            eq,
            "B*n",
            field=ToroidalMagneticField(1, 1),
            grid=LinearGrid(M=30, N=30, NFP=1, endpoint=True),
        )
        return fig


class TestPlotFSA:
    """Tests for plot_fsa."""

    @pytest.mark.unit
    def test_plot_fsa_axis_limit(self):
        """Test magnetic axis limit of flux surface average is plotted."""
        eq = get("W7-X")
        rho = np.linspace(0, 1, 10)
        grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, rho=rho)
        assert grid.axis.size

        name = "J*B"
        assert (
            "<" + name + ">" in data_index["desc.equilibrium.equilibrium.Equilibrium"]
        ), "Test with a different quantity."
        # should forward computation to compute function
        _, _, plot_data = plot_fsa(
            eq=eq,
            name=name,
            rho=rho,
            M=eq.M_grid,
            N=eq.N_grid,
            with_sqrt_g=True,
            return_data=True,
        )
        desired = grid.compress(
            eq.compute(names="<" + name + ">", grid=grid)["<" + name + ">"]
        )
        np.testing.assert_allclose(
            plot_data["<" + name + ">"], desired, equal_nan=False
        )

        name = "psi_r/sqrt(g)"
        assert (
            "<" + name + ">"
            not in data_index["desc.equilibrium.equilibrium.Equilibrium"]
        ), "Test with a different quantity."
        # should automatically compute axis limit
        _, _, plot_data = plot_fsa(
            eq=eq,
            name=name,
            rho=rho,
            M=eq.M_grid,
            N=eq.N_grid,
            with_sqrt_g=True,
            return_data=True,
        )
        data = eq.compute(names=[name, "sqrt(g)", "sqrt(g)_r"], grid=grid)
        desired = surface_averages(
            grid=grid,
            q=data[name],
            sqrt_g=grid.replace_at_axis(data["sqrt(g)"], data["sqrt(g)_r"], copy=True),
            expand_out=False,
        )
        np.testing.assert_allclose(
            plot_data["<" + name + ">_fsa"], desired, equal_nan=False
        )

        name = "|B|"
        assert (
            "<" + name + ">" in data_index["desc.equilibrium.equilibrium.Equilibrium"]
        ), "Test with a different quantity."
        _, _, plot_data = plot_fsa(
            eq=eq,
            name=name,
            rho=rho,
            M=eq.M_grid,
            N=eq.N_grid,
            with_sqrt_g=False,  # test that does not compute data_index["<|B|>"]
            return_data=True,
        )
        data = eq.compute(names=name, grid=grid)
        desired = surface_averages(grid=grid, q=data[name], expand_out=False)
        np.testing.assert_allclose(
            plot_data["<" + name + ">_fsa"], desired, equal_nan=False
        )

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_fsa_I(self):
        """Test plotting of flux surface average toroidal current."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_fsa(eq, "B_theta", with_sqrt_g=False, return_data=True)
        assert "rho" in data.keys()
        assert "<B_theta>_fsa" in data.keys()
        assert "normalization" in data.keys()
        assert data["normalization"] == 1

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_fsa_F_normalized(self):
        """Test plotting flux surface average normalized force error on log scale."""
        eq = get("DSHAPE_CURRENT")
        fig, ax = plot_fsa(
            eq, "|F|", log=True, norm_F=True, norm_name="<|grad(p)|>_vol"
        )
        ax.set_ylim([1e-6, 1e-3])
        return fig


class TestPlotSection:
    """Tests for plot_section."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_section_J(self):
        """Test plotting poincare section of radial current."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_section(eq, "J^rho", return_data=True)
        assert "R" in data.keys()
        assert "Z" in data.keys()
        assert "J^rho" in data.keys()
        assert "normalization" in data.keys()
        assert data["normalization"] == 1

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_section_chi_contour(self):
        """Test plotting poincare section of poloidal flux, with fill=False."""
        eq = get("DSHAPE_CURRENT")
        fig, ax = plot_section(eq, "chi", fill=False, levels=20)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_section_F(self):
        """Test plotting poincare section of radial force."""
        eq = get("DSHAPE_CURRENT")
        fig, ax = plot_section(eq, "F_rho")
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=50)
    def test_section_logF(self):
        """Test plotting poincare section of force magnitude on log scale."""
        eq = get("DSHAPE_CURRENT")
        fig, ax = plot_section(eq, "|F|", log=True)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_normF_section(self):
        """Test Poincare section plot of normalized force on log scale."""
        eq = get("DSHAPE_CURRENT")
        fig, ax = plot_section(eq, "|F|", norm_F=True, log=True)
        return fig


class TestPlotSurfaces:
    """Tests for plot_surfaces."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_surfaces(self):
        """Test plotting flux surfaces."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_surfaces(eq, return_data=True)
        for string in [
            "rho_R_coords",
            "rho_Z_coords",
            "vartheta_R_coords",
            "vartheta_Z_coords",
        ]:
            assert string in data.keys()

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_surfaces_no_theta(self):
        """Test plotting flux surfaces without theta contours."""
        eq = get("DSHAPE_CURRENT")
        fig, ax, data = plot_surfaces(eq, theta=False, return_data=True)
        for string in ["rho_R_coords", "rho_Z_coords"]:
            assert string in data.keys()

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_surfaces_HELIOTRON(self):
        """Test plot surfaces of equilibrium for correctness of vartheta lines."""
        fig, ax = plot_surfaces(get("HELIOTRON"))
        return fig


class TestPlotBoundary:
    """Tests for plot_boundary and plot_boundaries."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_boundary(self):
        """Test plotting boundary."""
        eq = get("W7-X")
        fig, ax, data = plot_boundary(eq, plot_axis=True, return_data=True)
        assert "R" in data.keys()
        assert "Z" in data.keys()

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boundary_surface(self):
        """Test plot_boundary function for Surface objects."""
        surf = FourierRZToroidalSurface()
        fig, ax = plot_boundary(surf)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boundaries(self):
        """Test plotting boundaries."""
        eq1 = get("SOLOVEV")
        eq2 = get("DSHAPE")
        eq3 = get("W7-X")
        eq4 = get("ESTELL")
        with pytest.raises(ValueError, match="differing field periods"):
            fig, ax = plot_boundaries([eq3, eq4], theta=0)
        _, _, data1 = plot_boundaries(
            (eq1, eq2, eq3),
            phi=4,
            return_data=True,
        )
        fig, ax, data = plot_boundaries(
            (eq1, eq2, eq3),
            phi=np.linspace(0, 2 * np.pi / eq3.NFP, 4, endpoint=False),
            return_data=True,
        )
        assert "R" in data.keys()
        assert "Z" in data.keys()
        assert len(data["R"]) == 3
        assert len(data["Z"]) == 3
        assert (
            data["R"][-1].shape == data1["R"][-1].shape
        ), "Passing phi as an integer or array results in different behavior"
        assert (
            data["Z"][-1].shape == data1["Z"][-1].shape
        ), "Passing phi as an integer or array results in different behavior"

        return fig


class TestPlotComparison:
    """Tests for plot_comparison."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_comparison(self):
        """Test plotting comparison of flux surfaces."""
        eqf = get("DSHAPE_CURRENT", "all")
        fig, ax, data = plot_comparison(eqf, return_data=True)
        for string in [
            "rho_R_coords",
            "rho_Z_coords",
            "vartheta_R_coords",
            "vartheta_Z_coords",
        ]:
            assert string in data.keys()
            assert len(data[string]) == len(eqf)

        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_comparison_no_theta(self):
        """Test plotting comparison of flux surfaces without theta contours."""
        eqf = get("DSHAPE_CURRENT", "all")
        fig, ax = plot_comparison(eqf, theta=0)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_comparison_different_NFPs(self):
        """Test plotting comparison of flux surfaces with differing NFPs."""
        eq = get("SOLOVEV")
        eq_nonax = get("HELIOTRON")
        eq_nonax2 = get("ESTELL")
        with pytest.raises(ValueError, match="differing field periods"):
            fig, ax = plot_comparison([eq_nonax, eq_nonax2], theta=0)
        fig, ax = plot_comparison(
            [eq, eq_nonax],
            phi=np.linspace(0, 2 * np.pi / eq_nonax.NFP, 6, endpoint=False),
            theta=0,
        )
        return fig


class TestPlotGrid:
    """Tests for the plot_grid function."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_linear(self):
        """Test plotting linear grid."""
        grid = LinearGrid(rho=10, theta=10, zeta=1)
        fig, ax, data = plot_grid(grid, return_data=True)
        for string in ["theta", "rho"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_quad(self):
        """Test plotting quadrature grid."""
        grid = QuadratureGrid(L=10, M=10, N=1)
        fig, ax = plot_grid(grid, figsize=(6, 6))
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_jacobi(self):
        """Test plotting concentric grid with jacobi nodes."""
        grid = ConcentricGrid(L=20, M=10, N=1, node_pattern="jacobi")
        fig, ax = plot_grid(grid)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_cheb1(self):
        """Test plotting concentric grid with chebyshev 1 nodes."""
        grid = ConcentricGrid(L=20, M=10, N=1, node_pattern="cheb1")
        fig, ax = plot_grid(grid)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_cheb2(self):
        """Test plotting concentric grid with chebyshev 2 nodes."""
        grid = ConcentricGrid(L=20, M=10, N=1, node_pattern="cheb2")
        fig, ax = plot_grid(grid)
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_grid_ocs(self):
        """Test plotting concentric grid with optimal concentric sampling nodes."""
        grid = ConcentricGrid(L=20, M=10, N=1, node_pattern="ocs")
        fig, ax = plot_grid(grid)
        return fig


class TestPlotBasis:
    """Tests for plot_basis function."""

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_basis_powerseries(self):
        """Test plotting power series basis."""
        basis = PowerSeries(L=6)
        fig, ax, data = plot_basis(basis, return_data=True)
        for string in ["amplitude", "rho", "l"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_basis_fourierseries(self):
        """Test plotting fourier series basis."""
        basis = FourierSeries(N=3)
        fig, ax, data = plot_basis(basis, return_data=True)
        for string in ["amplitude", "n", "zeta"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_basis_doublefourierseries(self):
        """Test plotting double fourier series basis."""
        basis = DoubleFourierSeries(M=3, N=2)
        fig, ax, data = plot_basis(basis, return_data=True)
        for string in ["amplitude", "n", "zeta", "m", "theta"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=26)
    def test_plot_basis_fourierzernike(self):
        """Test plotting fourier-zernike basis."""
        basis = FourierZernikeBasis(L=8, M=3, N=2)
        fig, ax, data = plot_basis(basis, return_data=True)
        for string in ["amplitude", "l", "rho", "m", "theta"]:
            assert string in data.keys()
        return fig


class TestPlotBoozerModes:
    """Tests for plot_boozer_modes."""

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boozer_modes(self):
        """Test plotting boozer spectrum."""
        eq = get("WISTELL-A")
        fig, ax, data = plot_boozer_modes(
            eq,
            M_booz=eq.M,
            N_booz=eq.N,
            num_modes=7,
            return_data=True,
            norm=True,
            rho=5,
        )
        ax.set_ylim([1e-6, 5e0])
        for string in ["|B|_mn", "B modes", "rho"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boozer_modes_breaking_only(self):
        """Test plotting symmetry breaking boozer spectrum."""
        eq = get("WISTELL-A")
        fig, ax = plot_boozer_modes(
            eq,
            M_booz=eq.M,
            N_booz=eq.N,
            helicity=(1, -eq.NFP),
            norm=True,
            num_modes=7,
            rho=5,
        )
        ax.set_ylim([1e-6, 5e0])
        return fig

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boozer_modes_max(self):
        """Test plotting symmetry breaking boozer spectrum."""
        eq = get("WISTELL-A")
        fig, ax = plot_boozer_modes(
            eq,
            M_booz=eq.M,
            N_booz=eq.N,
            helicity=(1, -eq.NFP),
            max_only=True,
            label="WISTELL-A",
            color="r",
            norm=True,
            rho=5,
        )
        ax.set_ylim([1e-6, 5e0])
        return fig

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
    def test_plot_boozer_modes_no_norm(self):
        """Test plotting boozer spectrum without B0 and norm."""
        eq = get("ESTELL")
        fig, ax = plot_boozer_modes(
            eq,
            M_booz=eq.M,
            N_booz=eq.N,
            num_modes=7,
            B0=False,
            log=False,
            rho=5,
        )
        return fig


class TestPlotBoozerSurface:
    """Tests for plot_boozer_surface."""

    @pytest.mark.unit
    @pytest.mark.slow
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_boozer_surface(self):
        """Test plotting B in boozer coordinates."""
        eq = get("WISTELL-A")
        fig, ax, data = plot_boozer_surface(
            eq, M_booz=eq.M, N_booz=eq.N, return_data=True, rho=0.5, fieldlines=4
        )
        for string in ["|B|", "theta_B", "zeta_B"]:
            assert string in data.keys()
        return fig

    @pytest.mark.unit
    @pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
    def test_plot_omnigenous_field(self):
        """Test plot omnigenous magnetic field."""
        field = OmnigenousField(
            L_B=0,
            M_B=4,
            L_x=0,
            M_x=1,
            N_x=1,
            NFP=4,
            helicity=(1, 4),
            B_lm=np.array([0.8, 0.9, 1.1, 1.2]),
            x_lmn=np.array([0, -np.pi / 8, 0, np.pi / 8, 0, np.pi / 4]),
        )
        fig, ax = plot_boozer_surface(field, iota=0.6, fieldlines=4)
        return fig


@pytest.mark.unit
@pytest.mark.slow
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
def test_plot_qs_error():
    """Test plotting qs error metrics."""
    eq = get("WISTELL-A")
    fig, ax, data = plot_qs_error(
        eq,
        helicity=(1, -eq.NFP),
        M_booz=eq.M,
        N_booz=eq.N,
        log=True,
        return_data=True,
        rho=5,
    )
    ax.set_ylim([1e-3, 2e-1])
    for string in ["rho", "f_T", "f_B", "f_C"]:
        assert string in data.keys()
        if string != "rho":
            # ensure that there is different QS data for each surface
            # related to gh PR #400
            assert not isalmostequal(data[string])
    return fig


@pytest.mark.unit
def test_plot_coils():
    """Test 3d plotting of coils with currents."""
    N = 48
    NFP = 4
    I = 1
    coil = FourierXYZCoil()
    coil.rotate(angle=np.pi / N)
    coils = CoilSet.linspaced_angular(coil, I, [0, 0, 1], np.pi / NFP, N // NFP // 2)
    coils2 = MixedCoilSet.from_symmetry(coils, NFP, True)
    with pytest.raises(ValueError, match="Expected `coils`"):
        plot_coils("not a coil")
    fig, data = plot_coils(coils2, return_data=True)

    def flatten_coils(coilset):
        if hasattr(coilset, "__len__"):
            return [a for i in coilset for a in flatten_coils(i)]
        else:
            return [coilset]

    coil_list = flatten_coils(coils2)
    for string in ["X", "Y", "Z"]:
        assert string in data.keys()
        assert len(data[string]) == len(coil_list)
    return fig


@pytest.mark.unit
def test_plot_coils_no_grid():
    """Test 3d plotting of coils with currents without any gridlines."""
    N = 48
    NFP = 4
    I = 1
    coil = FourierXYZCoil()
    coil.rotate(angle=np.pi / N)
    coils = CoilSet.linspaced_angular(coil, I, [0, 0, 1], np.pi / NFP, N // NFP // 2)
    with pytest.raises(ValueError, match="Expected `coils`"):
        plot_coils("not a coil")
    fig, data = plot_coils(
        coils,
        unique=True,
        return_data=True,
        showgrid=False,
        zeroline=False,
        showticklabels=False,
        showaxislabels=False,
    )

    def flatten_coils(coilset):
        if hasattr(coilset, "__len__"):
            return [a for i in coilset for a in flatten_coils(i)]
        else:
            return [coilset]

    coil_list = flatten_coils(coils)
    for string in ["X", "Y", "Z"]:
        assert string in data.keys()
        assert len(data[string]) == len(coil_list)
    return fig


@pytest.mark.unit
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_1d)
def test_plot_b_mag():
    """Test plot of |B| on longer field lines for gyrokinetic simulations."""
    psi = 0.5
    rho = np.sqrt(psi)
    npol = 2
    nzgrid = 128
    alpha = 0
    # compute iota
    eq = get("W7-X")
    iota = eq.compute("iota", grid=LinearGrid(rho=rho, NFP=eq.NFP))["iota"][0]

    # get flux tube coordinate system
    zeta = np.linspace(
        -np.pi * npol / np.abs(iota), np.pi * npol / np.abs(iota), 2 * nzgrid + 1
    )
    thetas = alpha * np.ones_like(zeta) + iota * zeta

    rhoa = rho * np.ones_like(zeta)
    c = np.vstack([rhoa, thetas, zeta]).T
    coords = eq.map_coordinates(c, inbasis=("rho", "theta_PEST", "zeta"))
    grid = Grid(coords)

    # compute |B| normalized in the usual flux tube way
    psib = np.abs(eq.compute("psi")["psi"][-1])
    Lref = eq.compute("a")["a"]
    Bref = 2 * psib / Lref**2
    bmag = eq.compute("|B|", grid=grid)["|B|"] / Bref
    fig, ax = plt.subplots()
    ax.plot(bmag)
    return fig


@pytest.mark.unit
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
def test_plot_coefficients():
    """Test scatter plot of spectral coefficients."""
    eq = get("DSHAPE_CURRENT")
    fig, ax = plot_coefficients(eq, color="b", marker="o")
    ax[0, 0].set_ylim([1e-8, 1e1])
    return fig


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
@pytest.mark.unit
def test_plot_logo():
    """Test plotting the DESC logo."""
    fig, ax = plot_logo()
    return fig


@pytest.mark.unit
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=tol_2d)
def test_plot_poincare():
    """Test making a poincare plot."""
    ext_field = load("tests/inputs/precise_QA_helical_coils.h5")
    eq = get("precise_QA")
    grid_trace = LinearGrid(rho=np.linspace(0, 1, 9))
    r0 = eq.compute("R", grid=grid_trace)["R"]
    z0 = eq.compute("Z", grid=grid_trace)["Z"]

    fig, ax = poincare_plot(ext_field, r0, z0, ntransit=50, NFP=eq.NFP)
    return fig



"""Magnetic field due to sheet current on a winding surface."""

import warnings

import matplotlib.pyplot as plt
import numpy as np
import skimage.measure
from scipy.constants import mu_0

from desc.backend import cho_factor, cho_solve, fori_loop, jnp
from desc.basis import DoubleFourierSeries
from desc.compute import rpz2xyz, rpz2xyz_vec, xyz2rpz_vec
from desc.compute.utils import _compute as compute_fun
from desc.derivatives import Derivative
from desc.geometry import FourierRZToroidalSurface
from desc.grid import Grid, LinearGrid
from desc.integrals import compute_B_plasma
from desc.optimizable import optimizable_parameter
from desc.utils import (
    Timer,
    check_posint,
    copy_coeffs,
    dot,
    errorif,
    safediv,
    setdefault,
    warnif,
)

from ._core import (
    _MagneticField,
    biot_savart_general,
    biot_savart_general_vector_potential,
)


class CurrentPotentialField(_MagneticField, FourierRZToroidalSurface):
    """Magnetic field due to a surface current potential on a toroidal surface.

    Surface current K is assumed given by K = n x ∇ Φ
    where:

        - n is the winding surface unit normal.
        - Phi is the current potential function, which is a function of theta and zeta.

    This function then uses biot-savart to find the B field from this current
    density K on the surface.

    Parameters
    ----------
    potential : callable
        function to compute the current potential. Should have a signature of
        the form potential(theta,zeta,**params) -> ndarray.
        theta,zeta are poloidal and toroidal angles on the surface.
        Assumed to have units of Amperes.
    potential_dtheta: callable
        function to compute the theta derivative of the current potential
    potential_dzeta: callable
        function to compute the zeta derivative of the current potential
    params : dict, optional
        default parameters to pass to potential function (and its derivatives)
    R_lmn, Z_lmn : array-like, shape(k,)
        Fourier coefficients for winding surface R and Z in cylindrical coordinates
    modes_R : array-like, shape(k,2)
        poloidal and toroidal mode numbers [m,n] for R_lmn.
    modes_Z : array-like, shape(k,2)
        mode numbers associated with Z_lmn, defaults to modes_R
    NFP : int
        number of field periods
    sym : bool
        whether to enforce stellarator symmetry for the surface geometry.
        Default is "auto" which enforces if modes are symmetric. If True,
        non-symmetric modes will be truncated.
    M, N: int or None
        Maximum poloidal and toroidal mode numbers. Defaults to maximum from modes_R
        and modes_Z.
    name : str
        name for this field
    check_orientation : bool
        ensure that this surface has a right handed orientation. Do not set to False
        unless you are sure the parameterization you have given is right handed
        (ie, e_theta x e_zeta points outward from the surface).

    """

    _io_attrs_ = (
        _MagneticField._io_attrs_
        + FourierRZToroidalSurface._io_attrs_
        + [
            "_params",
        ]
    )

    def __init__(
        self,
        potential,
        potential_dtheta,
        potential_dzeta,
        params=None,
        R_lmn=None,
        Z_lmn=None,
        modes_R=None,
        modes_Z=None,
        NFP=1,
        sym="auto",
        M=None,
        N=None,
        name="",
        check_orientation=True,
    ):
        assert callable(potential), "Potential must be callable!"
        assert callable(potential_dtheta), "Potential derivative must be callable!"
        assert callable(potential_dzeta), "Potential derivative must be callable!"

        self._potential = potential
        self._potential_dtheta = potential_dtheta
        self._potential_dzeta = potential_dzeta
        self._params = params

        super().__init__(
            R_lmn=R_lmn,
            Z_lmn=Z_lmn,
            modes_R=modes_R,
            modes_Z=modes_Z,
            NFP=NFP,
            sym=sym,
            M=M,
            N=N,
            name=name,
            check_orientation=check_orientation,
        )

    # TODO: make this an optimizable parameter so the potential may be optimized
    @property
    def params(self):
        """Dict of parameters to pass to potential function and its derivatives."""
        return self._params

    @params.setter
    def params(self, new):
        warnif(
            len(new) != len(self._params),
            UserWarning,
            "Length of new params is different from length of current params! "
            "May cause errors unless potential function is also changed.",
        )
        self._params = new

    @property
    def potential(self):
        """Potential function, signature (theta,zeta,**params) -> potential value."""
        return self._potential

    @potential.setter
    def potential(self, new):
        if new != self._potential:
            assert callable(new), "Potential must be callable!"
            self._potential = new

    @property
    def potential_dtheta(self):
        """Phi poloidal deriv. function, signature (theta,zeta,**params) -> value."""
        return self._potential_dtheta

    @potential_dtheta.setter
    def potential_dtheta(self, new):
        if new != self._potential_dtheta:
            assert callable(new), "Potential derivative must be callable!"
            self._potential_dtheta = new

    @property
    def potential_dzeta(self):
        """Phi toroidal deriv. function, signature (theta,zeta,**params) -> value."""
        return self._potential_dzeta

    @potential_dzeta.setter
    def potential_dzeta(self, new):
        if new != self._potential_dzeta:
            assert callable(new), "Potential derivative must be callable!"
            self._potential_dzeta = new

    def save(self, file_name, file_format=None, file_mode="w"):
        """Save the object.

        **Not supported for this object!**

        Parameters
        ----------
        file_name : str file path OR file instance
            location to save object
        file_format : str (Default hdf5)
            format of save file. Only used if file_name is a file path
        file_mode : str (Default w - overwrite)
            mode for save file. Only used if file_name is a file path

        """
        raise OSError(
            "Saving CurrentPotentialField is not supported,"
            " as the potential function cannot be serialized."
        )

    def _compute_A_or_B(
        self,
        coords,
        params=None,
        basis="rpz",
        source_grid=None,
        transforms=None,
        compute_A_or_B="B",
    ):
        """Compute magnetic field or vector potential at a set of points.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate field at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic field.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid
        compute_A_or_B: {"A", "B"}, optional
            whether to compute the magnetic vector potential "A" or the magnetic field
            "B". Defaults to "B"

        Returns
        -------
        field : ndarray, shape(N,3)
            magnetic field or vector potential at specified points

        """
        source_grid = source_grid or LinearGrid(
            M=30 + 2 * self.M,
            N=30 + 2 * self.N,
            NFP=self.NFP,
        )
        return _compute_A_or_B_from_CurrentPotentialField(
            field=self,
            coords=coords,
            params=params,
            basis=basis,
            source_grid=source_grid,
            transforms=transforms,
            compute_A_or_B=compute_A_or_B,
        )

    def compute_magnetic_field(
        self, coords, params=None, basis="rpz", source_grid=None, transforms=None
    ):
        """Compute magnetic field at a set of points.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate field at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic field.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid

        Returns
        -------
        field : ndarray, shape(N,3)
            magnetic field at specified points

        """
        return self._compute_A_or_B(coords, params, basis, source_grid, transforms, "B")

    def compute_magnetic_vector_potential(
        self, coords, params=None, basis="rpz", source_grid=None, transforms=None
    ):
        """Compute magnetic vector potential at a set of points.

        This assumes the Coulomb gauge.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate vector potential at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic vector potential.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid

        Returns
        -------
        A : ndarray, shape(N,3)
            Magnetic vector potential at specified points.

        """
        return self._compute_A_or_B(coords, params, basis, source_grid, transforms, "A")

    @classmethod
    def from_surface(
        cls,
        surface,
        potential,
        potential_dtheta,
        potential_dzeta,
        params=None,
    ):
        """Create CurrentPotentialField using geometry provided by given surface.

        Parameters
        ----------
        surface: FourierRZToroidalSurface, optional, default None
            Existing FourierRZToroidalSurface object to create a
            CurrentPotentialField with.
        potential : callable
            function to compute the current potential. Should have a signature of
            the form potential(theta,zeta,**params) -> ndarray.
            theta,zeta are poloidal and toroidal angles on the surface
        potential_dtheta: callable
            function to compute the theta derivative of the current potential
        potential_dzeta: callable
            function to compute the zeta derivative of the current potential
        params : dict, optional
            default parameters to pass to potential function (and its derivatives)

        """
        errorif(
            not isinstance(surface, FourierRZToroidalSurface),
            TypeError,
            "Expected type FourierRZToroidalSurface for argument surface, "
            f"instead got type {type(surface)}",
        )

        R_lmn = surface.R_lmn
        Z_lmn = surface.Z_lmn
        modes_R = surface._R_basis.modes[:, 1:]
        modes_Z = surface._Z_basis.modes[:, 1:]
        NFP = surface.NFP
        sym = surface.sym
        name = surface.name

        return cls(
            potential,
            potential_dtheta,
            potential_dzeta,
            params,
            R_lmn,
            Z_lmn,
            modes_R,
            modes_Z,
            NFP,
            sym,
            name=name,
            check_orientation=False,
        )


class FourierCurrentPotentialField(_MagneticField, FourierRZToroidalSurface):
    """Magnetic field due to a surface current potential on a toroidal surface.

    Surface current K is assumed given by

    K = n x ∇ Φ

    Φ(θ,ζ) = Φₛᵥ(θ,ζ) + Gζ/2π + Iθ/2π

    where:

        - n is the winding surface unit normal.
        - Phi is the current potential function, which is a function of theta and zeta,
          and is given as a secular linear term in theta/zeta and a double Fourier
          series in theta/zeta.

    This class then uses biot-savart to find the B field from this current
    density K on the surface.

    Parameters
    ----------
    Phi_mn : ndarray
        Fourier coefficients of the double FourierSeries part of the current potential.
        Has units of Amperes.
    modes_Phi : array-like, shape(k,2)
        Poloidal and Toroidal mode numbers corresponding to passed-in Phi_mn
        coefficients.
    I : float
        Net current linking the plasma and the surface toroidally
        Denoted I in the algorithm, has units of Amperes.
    G : float
        Net current linking the plasma and the surface poloidally
        Denoted G in the algorithm, has units of Amperes.
        NOTE: a negative G will tend to produce a positive toroidal magnetic field
        B in DESC, as in DESC the poloidal angle is taken to be positive
        and increasing when going in the clockwise direction, which with the
        convention n x grad(phi) will result in a toroidal field in the negative
        toroidal direction.
    sym_Phi :  {False,"cos","sin"}
        whether to enforce a given symmetry for the DoubleFourierSeries part of the
        current potential.
    M_Phi, N_Phi: int or None
        Maximum poloidal and toroidal mode numbers for the single valued part of the
        current potential.
    R_lmn, Z_lmn : array-like, shape(k,)
        Fourier coefficients for winding surface R and Z in cylindrical coordinates
    modes_R : array-like, shape(k,2)
        poloidal and toroidal mode numbers [m,n] for R_lmn.
    modes_Z : array-like, shape(k,2)
        mode numbers associated with Z_lmn, defaults to modes_R
    NFP : int
        number of field periods
    sym : bool
        whether to enforce stellarator symmetry for the surface geometry.
        Default is "auto" which enforces if modes are symmetric. If True,
        non-symmetric modes will be truncated.
    M, N: int or None
        Maximum poloidal and toroidal mode numbers. Defaults to maximum from modes_R
        and modes_Z.
    name : str
        name for this field
    check_orientation : bool
        ensure that this surface has a right handed orientation. Do not set to False
        unless you are sure the parameterization you have given is right handed
        (ie, e_theta x e_zeta points outward from the surface).

    """

    _io_attrs_ = (
        _MagneticField._io_attrs_
        + FourierRZToroidalSurface._io_attrs_
        + ["_Phi_mn", "_I", "_G", "_Phi_basis", "_M_Phi", "_N_Phi", "_sym_Phi"]
    )

    def __init__(
        self,
        Phi_mn=np.array([0.0]),
        modes_Phi=np.array([[0, 0]]),
        I=0,
        G=0,
        sym_Phi=False,
        M_Phi=None,
        N_Phi=None,
        R_lmn=None,
        Z_lmn=None,
        modes_R=None,
        modes_Z=None,
        NFP=1,
        sym="auto",
        M=None,
        N=None,
        name="",
        check_orientation=True,
    ):
        Phi_mn, modes_Phi = map(np.asarray, (Phi_mn, modes_Phi))
        assert (
            Phi_mn.size == modes_Phi.shape[0]
        ), "Phi_mn size and modes_Phi.shape[0] must be the same size!"

        assert np.issubdtype(modes_Phi.dtype, np.integer)

        M_Phi = setdefault(M_Phi, np.max(abs(modes_Phi[:, 0])))
        N_Phi = setdefault(N_Phi, np.max(abs(modes_Phi[:, 1])))

        self._M_Phi = M_Phi
        self._N_Phi = N_Phi

        self._sym_Phi = sym_Phi
        self._Phi_basis = DoubleFourierSeries(M=M_Phi, N=N_Phi, NFP=NFP, sym=sym_Phi)
        self._Phi_mn = copy_coeffs(Phi_mn, modes_Phi, self._Phi_basis.modes[:, 1:])

        self._I = float(np.squeeze(I))
        self._G = float(np.squeeze(G))

        super().__init__(
            R_lmn=R_lmn,
            Z_lmn=Z_lmn,
            modes_R=modes_R,
            modes_Z=modes_Z,
            NFP=NFP,
            sym=sym,
            M=M,
            N=N,
            name=name,
            check_orientation=check_orientation,
        )

    @optimizable_parameter
    @property
    def I(self):  # noqa: E743
        """Net current linking the plasma and the surface toroidally."""
        return self._I

    @I.setter
    def I(self, new):  # noqa: E743
        self._I = float(np.squeeze(new))

    @optimizable_parameter
    @property
    def G(self):
        """Net current linking the plasma and the surface poloidally."""
        return self._G

    @G.setter
    def G(self, new):
        self._G = float(np.squeeze(new))

    @optimizable_parameter
    @property
    def Phi_mn(self):
        """Fourier coefficients describing single-valued part of potential."""
        return self._Phi_mn

    @Phi_mn.setter
    def Phi_mn(self, new):
        if len(new) == self.Phi_basis.num_modes:
            self._Phi_mn = jnp.asarray(new)
        else:
            raise ValueError(
                f"Phi_mn should have the same size as the basis, got {len(new)} for "
                + f"basis with {self.Phi_basis.num_modes} modes."
            )

    @property
    def Phi_basis(self):
        """DoubleFourierSeries: Spectral basis for Phi."""
        return self._Phi_basis

    @property
    def sym_Phi(self):
        """str: Type of symmetry of periodic part of Phi (no symmetry if False)."""
        return self._sym_Phi

    @property
    def M_Phi(self):
        """int: Poloidal resolution of periodic part of Phi."""
        return self._M_Phi

    @property
    def N_Phi(self):
        """int: Toroidal resolution of periodic part of Phi."""
        return self._N_Phi

    def change_Phi_resolution(self, M=None, N=None, NFP=None, sym_Phi=None):
        """Change the maximum poloidal and toroidal resolution for Phi.

        Parameters
        ----------
        M : int
            Poloidal resolution to change Phi basis to.
            If None, defaults to current self.Phi_basis poloidal resolution
        N : int
            Toroidal resolution to change Phi basis to.
            If None, defaults to current self.Phi_basis toroidal resolution
        NFP : int
            Number of field periods for surface and Phi basis.
            If None, defaults to current NFP.
            Note: will change the NFP of the surface geometry as well as the
            Phi basis.
        sym_Phi :  {"auto","cos","sin",False}
            whether to enforce a given symmetry for the DoubleFourierSeries part of the
            current potential. Default is "auto" which enforces if modes are symmetric.
            If True, non-symmetric modes will be truncated.

        """
        M = M or self._M_Phi
        N = N or self._M_Phi
        NFP = NFP or self.NFP
        sym_Phi = sym_Phi or self.sym_Phi

        Phi_modes_old = self.Phi_basis.modes
        self.Phi_basis.change_resolution(M=M, N=N, NFP=self.NFP, sym=sym_Phi)

        self._Phi_mn = copy_coeffs(self.Phi_mn, Phi_modes_old, self.Phi_basis.modes)
        self._M_Phi = M
        self._N_Phi = N
        self._sym_Phi = sym_Phi
        self.change_resolution(
            NFP=NFP
        )  # make sure surface and Phi basis NFP are the same

    def _compute_A_or_B(
        self,
        coords,
        params=None,
        basis="rpz",
        source_grid=None,
        transforms=None,
        compute_A_or_B="B",
    ):
        """Compute magnetic field or vector potential at a set of points.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate field at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic field.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid
        compute_A_or_B: {"A", "B"}, optional
            whether to compute the magnetic vector potential "A" or the magnetic field
            "B". Defaults to "B"

        Returns
        -------
        field : ndarray, shape(N,3)
            magnetic field or vector potential at specified points

        """
        source_grid = source_grid or LinearGrid(
            M=30 + 2 * max(self.M, self.M_Phi),
            N=30 + 2 * max(self.N, self.N_Phi),
            NFP=self.NFP,
        )
        return _compute_A_or_B_from_CurrentPotentialField(
            field=self,
            coords=coords,
            params=params,
            basis=basis,
            source_grid=source_grid,
            transforms=transforms,
            compute_A_or_B=compute_A_or_B,
        )

    def compute_magnetic_field(
        self, coords, params=None, basis="rpz", source_grid=None, transforms=None
    ):
        """Compute magnetic field at a set of points.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate field at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic field.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid

        Returns
        -------
        field : ndarray, shape(N,3)
            magnetic field at specified points

        """
        return self._compute_A_or_B(coords, params, basis, source_grid, transforms, "B")

    def compute_magnetic_vector_potential(
        self, coords, params=None, basis="rpz", source_grid=None, transforms=None
    ):
        """Compute magnetic vector potential at a set of points.

        This assumes the Coulomb gauge.

        Parameters
        ----------
        coords : array-like shape(n,3)
            Nodes to evaluate vector potential at in [R,phi,Z] or [X,Y,Z] coordinates.
        params : dict or array-like of dict, optional
            Dictionary of optimizable parameters, eg field.params_dict.
        basis : {"rpz", "xyz"}
            Basis for input coordinates and returned magnetic vector potential.
        source_grid : Grid, int or None or array-like, optional
            Source grid upon which to evaluate the surface current density K.
        transforms : dict of Transform
            Transforms for R, Z, lambda, etc. Default is to build from source_grid

        Returns
        -------
        A : ndarray, shape(N,3)
            Magnetic vector potential at specified points.

        """
        return self._compute_A_or_B(coords, params, basis, source_grid, transforms, "A")

    @classmethod
    def from_surface(
        cls,
        surface,
        Phi_mn=np.array([0.0]),
        modes_Phi=np.array([[0, 0]]),
        I=0,
        G=0,
        sym_Phi="auto",
        M_Phi=None,
        N_Phi=None,
    ):
        """Create FourierCurrentPotentialField using geometry of given surface.

        Parameters
        ----------
        surface: FourierRZToroidalSurface, optional, default None
            Existing FourierRZToroidalSurface object to create a
            CurrentPotentialField with.
        Phi_mn : ndarray
            Fourier coefficients of the double FourierSeries of the current potential.
            Should correspond to the given DoubleFourierSeries basis object passed in.
        modes_Phi : array-like, shape(k,2)
            Poloidal and Toroidal mode numbers corresponding to passed-in Phi_mn
            coefficients
        I : float
            Net current linking the plasma and the surface toroidally
            Denoted I in the algorithm
        G : float
            Net current linking the plasma and the surface poloidally
            Denoted G in the algorithm
            NOTE: a negative G will tend to produce a positive toroidal magnetic field
            B in DESC, as in DESC the poloidal angle is taken to be positive
            and increasing when going in the clockwise direction, which with the
            convention n x grad(phi) will result in a toroidal field in the negative
            toroidal direction.
        sym_Phi :  {"auto", "cos","sin", False}
            whether to enforce a given symmetry for the DoubleFourierSeries part of the
            current potential. If "auto", assumes sin symmetry if the surface is
            symmetric, else False.
        M_Phi, N_Phi: int or None
            Maximum poloidal and toroidal mode numbers for the single valued part of the
            current potential.

        """
        if not isinstance(surface, FourierRZToroidalSurface):
            raise TypeError(
                "Expected type FourierRZToroidalSurface for argument surface, "
                f"instead got type {type(surface)}"
            )
        R_lmn = surface.R_lmn
        Z_lmn = surface.Z_lmn
        modes_R = surface._R_basis.modes[:, 1:]
        modes_Z = surface._Z_basis.modes[:, 1:]
        NFP = surface.NFP
        sym = surface.sym
        name = surface.name
        if sym_Phi == "auto":
            sym_Phi = "sin" if surface.sym else False

        return cls(
            Phi_mn=Phi_mn,
            modes_Phi=modes_Phi,
            I=I,
            G=G,
            sym_Phi=sym_Phi,
            M_Phi=M_Phi,
            N_Phi=N_Phi,
            R_lmn=R_lmn,
            Z_lmn=Z_lmn,
            modes_R=modes_R,
            modes_Z=modes_Z,
            NFP=NFP,
            sym=sym,
            name=name,
            check_orientation=False,
        )

    def to_CoilSet(  # noqa: C901 - FIXME: simplify this
        self,
        num_coils,
        step=1,
        spline_method="cubic",
        show_plots=False,
        npts=128,
        stell_sym=False,
        plot_kwargs={"figsize": (8, 6)},
    ):
        """Find helical or modular coils from this surface current potential.

        Surface current K is assumed given by

        K = n x ∇ Φ

        Φ(θ,ζ) = Φₛᵥ(θ,ζ) + Gζ/2π + Iθ/2π

        where n is the winding surface unit normal, Φ is the current potential
        function, which is a function of theta and zeta, and is given as a
        secular linear term in theta (I)  and zeta (G) and a double Fourier
        series in theta/zeta.

        NOTE: The function is not jit/AD compatible

        Parameters
        ----------
        num_coils : int, optional
            Number of coils to discretize the surface current with.
            If the coils are modular (i.e. I=0), then this is the number of
            coils per field period. If the coils are stellarator-symmetric, then this
            is the number of coils per half field-period. The coils returned always
            have a coil which passes through the theta=0 zeta=0 point of the surface.
        step : int, optional
            Amount of points to skip by when saving the coil geometry spline
            by default 1, meaning that every point will be saved
            if higher, less points will be saved e.g. 3 saves every 3rd point
        spline_method : str, optional
            method of fitting to use for the spline, by default ``"cubic"``
            see ``SplineXYZCoil`` for more info
        show_plots : bool, optional,
            whether to show plots of the contours chosen for coils, by default False
        npts : int, optional
            Number of zeta points over one field period to use to discretize the surface
            when finding constant current potential contours.
        stell_sym : bool
            whether the coils are stellarator-symmetric or not. Defaults to False. Only
            matters for modular coils (currently)
        plot_kwargs : dict
            dict of kwargs to use when plotting the contour plots if ``show_plots=True``
            ``figsize`` is used for the figure size, and the rest are passed to
            ``plt.contourf``

        Returns
        -------
        coils : CoilSet
            DESC `CoilSet` of `SplineXYZCoil` coils that are a discretization of
            the surface current on the given winding surface.

        """
        check_posint(num_coils, "num_coils", False)
        check_posint(step, "step", False)
        check_posint(npts, "npts", False)
        nfp = self.Phi_basis.NFP

        net_toroidal_current = self.I
        net_poloidal_current = self.G
        helicity = safediv(
            net_poloidal_current, net_toroidal_current * nfp, threshold=1e-8
        )
        coil_type = "modular" if jnp.isclose(helicity, 0) else "helical"
        # determine current per coil
        if coil_type == "helical":
            # helical coils
            coil_current = jnp.abs(net_toroidal_current) / num_coils
        else:  # modular coils
            coil_current = net_poloidal_current / num_coils / nfp
            if stell_sym:  # num_coils is num coils per half period, so
                # need to account for the extra factor of 2
                coil_current = coil_current / 2
        assert not jnp.isclose(net_toroidal_current, 0) or not jnp.isclose(
            net_poloidal_current, 0
        ), (
            "Detected both net toroidal and poloidal current are both zero, "
            "this function cannot find windowpane coils"
        )

        contour_theta, contour_zeta = _find_current_potential_contours(
            self,
            num_coils,
            npts,
            show_plots,
            stell_sym,
            net_poloidal_current,
            net_toroidal_current,
            helicity,
            plot_kwargs=plot_kwargs,
        )

        ################################################################
        # Find the XYZ points in real space of the coil contours
        ################################################################

        contour_X, contour_Y, contour_Z = _find_XYZ_points(
            contour_theta,
            contour_zeta,
            self,
        )
        ################################################################
        # Create CoilSet object
        ################################################################
        # local imports to avoid circular imports
        from desc.coils import CoilSet, SplineXYZCoil

        coils = []
        for j in range(num_coils):
            if coil_type == "helical":
                # helical coils
                # make sure that the sign of the coil current is correct
                # by dotting K with the vector along the contour
                # TODO: probably could use helicity sign and just check the slope of
                # the contours to see which way they are going, but this is easy for
                # now and not too expensive
                contour_vector = jnp.array(
                    [
                        contour_X[j][1] - contour_X[j][0],
                        contour_Y[j][1] - contour_Y[j][0],
                        contour_Z[j][1] - contour_Z[j][0],
                    ]
                )
                K = self.compute(
                    "K",
                    grid=Grid(
                        jnp.array([[0, contour_theta[j][0], contour_zeta[j][0]]])
                    ),
                    basis="xyz",
                )["K"]
                current_sign = jnp.sign(jnp.dot(contour_vector, K[0, :]))
                thisCurrent = current_sign * jnp.abs(coil_current)
            else:
                # modular coils
                # make sure that the sign of the coil current is correct
                # don't need to dot with K here because we know the direction
                # based off the direction of the theta contour and sign of G
                # (extra negative sign because a positive G -> negative toroidal B
                # but we always have a right-handed coord system, and so current flowing
                # in positive poloidal direction creates a positive toroidal B)
                # for modular coils, easiest way to check contour direction is to see
                # direction of the contour thetas
                sign_of_theta_contours = jnp.sign(
                    contour_theta[0][-1] - contour_theta[0][0]
                )
                current_sign = -sign_of_theta_contours * jnp.sign(net_poloidal_current)
                thisCurrent = jnp.abs(coil_current) * current_sign
            coil = SplineXYZCoil(
                thisCurrent,
                jnp.append(contour_X[j][0::step], contour_X[j][0]),
                jnp.append(contour_Y[j][0::step], contour_Y[j][0]),
                jnp.append(contour_Z[j][0::step], contour_Z[j][0]),
                method=spline_method,
            )
            coils.append(coil)
        # check_intersection is False here as these coils by construction
        # cannot intersect eachother (they are contours of the current potential
        # which cannot self-intersect by definition)
        # unless stell_sym is true, then the full coilset might have
        # self intersection depending on if the coils cross the
        # symmetry plane, in which case we will check
        if coil_type == "modular":
            final_coilset = CoilSet(
                *coils, NFP=nfp, sym=stell_sym, check_intersection=stell_sym
            )
        else:
            # TODO: once winding surface curve is implemented, enforce sym for
            # helical as well
            final_coilset = CoilSet(*coils, check_intersection=False)
        return final_coilset


def _compute_A_or_B_from_CurrentPotentialField(
    field,
    coords,
    source_grid,
    params=None,
    basis="rpz",
    transforms=None,
    compute_A_or_B="B",
):
    """Compute magnetic field or vector potential at a set of points.

    Parameters
    ----------
    field : CurrentPotentialField or FourierCurrentPotentialField
        current potential field object from which to compute magnetic field.
    coords : array-like shape(N,3)
        cylindrical or cartesian coordinates
    source_grid : Grid,
        source grid upon which to evaluate the surface current density K
    params : dict, optional
        parameters to pass to compute function
        should include the potential
    basis : {"rpz", "xyz"}
        basis for input coordinates and returned magnetic field
    compute_A_or_B: {"A", "B"}, optional
        whether to compute the magnetic vector potential "A" or the magnetic field
        "B". Defaults to "B"


    Returns
    -------
    field : ndarray, shape(N,3)
        magnetic field or vector potential at specified points

    """
    errorif(
        compute_A_or_B not in ["A", "B"],
        ValueError,
        f'Expected "A" or "B" for compute_A_or_B, instead got {compute_A_or_B}',
    )
    assert basis.lower() in ["rpz", "xyz"]
    coords = jnp.atleast_2d(jnp.asarray(coords))
    if basis == "rpz":
        coords = rpz2xyz(coords)
    op = {"B": biot_savart_general, "A": biot_savart_general_vector_potential}[
        compute_A_or_B
    ]
    # compute surface current, and store grid quantities
    # needed for integration in class
    if not params or not transforms:
        data = field.compute(
            ["K", "x"],
            grid=source_grid,
            basis="rpz",
            params=params,
            transforms=transforms,
            jitable=True,
        )
    else:
        data = compute_fun(
            field,
            names=["K", "x"],
            params=params,
            transforms=transforms,
            profiles={},
        )

    _rs = data["x"]
    _K = data["K"]

    # surface element, must divide by NFP to remove the NFP multiple on the
    # surface grid weights, as we account for that when doing the for loop
    # over NFP
    _dV = source_grid.weights * data["|e_theta x e_zeta|"] / source_grid.NFP

    def nfp_loop(j, f):
        # calculate (by rotating) rs, rs_t, rz_t
        phi = (source_grid.nodes[:, 2] + j * 2 * jnp.pi / source_grid.NFP) % (
            2 * jnp.pi
        )
        # new coords are just old R,Z at a new phi (bc of discrete NFP symmetry)
        rs = jnp.vstack((_rs[:, 0], phi, _rs[:, 2])).T
        rs = rpz2xyz(rs)
        K = rpz2xyz_vec(_K, phi=phi)
        fj = op(
            coords,
            rs,
            K,
            _dV,
        )
        f += fj
        return f

    B = fori_loop(0, source_grid.NFP, nfp_loop, jnp.zeros_like(coords))
    if basis == "rpz":
        B = xyz2rpz_vec(B, x=coords[:, 0], y=coords[:, 1])
    return B


def solve_regularized_surface_current(  # noqa: C901 fxn too complex
    field,
    eq,
    lambda_regularization=1e-30,
    current_helicity=(1, 0),
    vacuum=False,
    regularization_type="regcoil",
    source_grid=None,
    eval_grid=None,
    vc_source_grid=None,
    external_field=None,
    external_field_grid=None,
    verbose=1,
):
    """Runs REGCOIL-like algorithm to find the current potential for the surface.

    NOTE: The function is not jit/AD compatible

    Follows algorithm of [1]_ to find the current potential Phi on the surface,
    given a surface current::

        K = n x ∇ Φ
        Φ(θ,ζ) = Φₛᵥ(θ,ζ) + Gζ/2π + Iθ/2π

    The algorithm minimizes the quadratic flux on the plasma surface due to the
    surface current (B_Phi_SV for field from the single valued part Φₛᵥ, and
    B_GI for that from the secular terms I and G), plasma current, and external
    fields::

        Bn = ∫ ∫ (B . n)^2 dA
        B = B_plasma + B_external + B_Phi_SV + B_GI

    G is fixed by the equilibrium magnetic field strength, and I is determined
    by the desired coil topology (given by ``current_helicity``), with zero
    helicity corresponding to modular coils, and non-zero helicity corresponding
    to helical coils. The algorithm then finds the single-valued part of Φ
    by minimizing the quadratic flux on the plasma surface along with a
    regularization term on the surface current magnitude::

        min_Φₛᵥ  ∫ ∫ (B . n)^2 dA + λ ∫ ∫ ||K||^2 dA

    where λ is the regularization parameter, smaller `lambda_regularization`
    corresponds to less regularization (consequently, lower Bn error but more
    complex and large surface currents) and larger `lambda_regularization`
    corresponds to more regularization (consequently, higher Bn error but simpler
    and smaller surface currents).

    If the ``simple`` regularization is used, the problem instead becomes::

        min_Φₛᵥ  (B . n)^2 + λ  ||Φ_mn||^2

    Parameters
    ----------
    field : FourierCurrentPotentialField
        ``FourierCurrentPotentialField`` to run REGCOIL algorithm with.
    eq : Equilibrium
        Equilibrium to minimize the quadratic flux (plus regularization) on.
    lambda_regularization : float or ndarray, optional
        regularization parameter, >= 0, regularizes minimization of Bn
        on plasma surface with minimization of current density mag K on winding
        surface i.e. larger lambda_regularization, simpler coilset and smaller
        currents, but worse Bn. If a float, only runs REGCOIL for that single value
        and returns a list with the single FourierCurrentPotentialField and the
        associated data.
        If an array is passed, will run REGCOIL for each lambda_regularization in
        that array and return a list of FourierCurrentPotentialFields, and the
        associated data.
    current_helicity : tuple of size 2, optional
        Tuple of ``(M_coil, N_coil)`` used to determine coil topology, where`` M_coil``
        is the number of poloidal transits a coil makes before closing back on itself
        and ``N_coil`` is the number of toroidal transits a coil makes before
        returning back to itself.
        if ``N_coil`` is zero and ``M_coil`` nonzero, it corresponds to modular
        coil topology.
        If both ``N_coil``,``M_coil`` are nonzero, it corresponds to helical coils.
        If ``N_coil``,``M_coil`` are both zero, it corresponds to windowpane coils.
        The net toroidal current (when ``M_coil`` is nonzero) is set as
        ``I = N_coil(G-G_ext)/M_coil``
        As an example, if helical coils which make one poloidal transit per field period
        and close on themselves after one full toroidal transit are desired, that
        corresponds to ``current_helicity = (1*NFP, 1)``
    vacuum : bool, optional
        if True, will not include the contribution to the normal field from the
        plasma currents.
    regularization_type : {"simple","regcoil"}
        whether to use a simple regularization based off of just the single-valued
        part of Phi, or to use the full REGCOIL regularization penalizing | K | ^ 2.
        Defaults to ``"regcoil"``
    source_grid : Grid, optional
        Source grid upon which to evaluate the surface current when calculating
        the normal field on the plasma surface. Defaults to
        LinearGrid(M=max(3 * current_potential_field.M_Phi, 30),
        N=max(3 * current_potential_field.N_Phi, 30), NFP=eq.NFP)
    eval_grid : Grid, optional
        Grid upon which to evaluate the normal field on the plasma surface, and
        at which the normal field is minimized.
        Defaults to
        `LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP)`
    vc_source_grid : LinearGrid
        LinearGrid to use for the singular integral for the virtual casing
        principle to calculate the component of the normal field from the
        plasma currents. Must have endpoint=False and sym=False and be linearly
        spaced in theta and zeta, with nodes only at rho=1.0
    external_field: _MagneticField,
        DESC `_MagneticField` object giving the magnetic field
        provided by any coils/fields external to the winding surface.
        e.g. can provide a TF coilset to calculate the surface current
        which is needed to minimize Bn given this external coilset providing
        the bulk of the required net toroidal magnetic flux, by default None
    external_field_grid : Grid, optional
        Source grid with which to evaluate the external field when calculating
        its contribution to the normal field on the plasma surface (if it is a type
        that requires a source, like a `CoilSet` or a `CurrentPotentialField`).
        By default None, which will use the default grid for the given
        external field type.
    verbose : int, optional
        level of verbosity, if 0 will print nothing.
        1 will display Bn max,min,average and chi^2 values for each
        lambda_regularization.
        2 will display jacobian timing info

    Returns
    -------
    fields  : list of FourierCurrentPotentialField
        A FourierCurrentPotentialField with the Phi_mn set to the
        optimized current potential. This is a list of length
        lambda_regularization.size with the optimized fields
        for each parameter value lambda_regularization.
    data : dict
        Dictionary with the following keys,::

            lambda_regularization : regularization parameter the algorithm was ran
                    with, a array of passed-in `lambda_regularization`
                    corresponding to the list of `Phi_mn`.
            Phi_mn : the single-valued current potential coefficients which
                    minimize the Bn at the given eval_grid on the plasma, subject
                    to regularization on the surface current magnitude governed by
                    lambda_regularization.
                    A list of arrays of length `self.Phi_basis.num_modes` if passed-in
                    `lambda_regularization`,
                    with list length `lambda_regularization.size`, corresponding to the
                    list of regularization parameters `lambda_regularization`.
            I : float, net toroidal current (in Amperes) on the winding surface.
                    Governed by the `current_helicity` parameters, and is zero for
                    modular coils (when `p=current_helicity[0]=0`).
            G : float, net poloidal current (in Amperes) on the winding surface.
                    Determined by the equilibrium toroidal magnetic field, as well as
                    the given external field.
            chi^2_B : quadratic flux squared, integrated over the plasma surface.
                    list of float of  length `lambda_regularization.size`,
                    corresponding to the array of `lambda_regularization` values.
            chi^2_K : Current density magnitude squared, integrated over winding
                    surface. a list of float of length `lambda_regularization.size`,
                    corresponding to the array of `lambda_regularization`.
            ||K|| : Current density magnitude on winding surface, evaluated at the
                    given `source_grid`. A list of arrays, with list length
                    `lambda_regularization.size`, corresponding to the array
                    of `lambda_regularization`.
            eval_grid: Grid object that Bn was evaluated at.
            source_grid: Grid object that Phi and K were evaluated at.

    References
    ----------
    .. [1] Landreman, Matt. "An improved current potential method for fast computation
      of stellarator coil shapes." Nuclear Fusion 57 (2017): 046003.

    """
    errorif(
        len(current_helicity) != 2,
        ValueError,
        "current_helicity must be a length-two tuple",
    )
    errorif(
        any(int(hel) != hel for hel in current_helicity),
        ValueError,
        "Helicity values must be integer",
    )

    errorif(
        regularization_type not in ["simple", "regcoil"],
        ValueError,
        'regularization_type must be "simple" or "regcoil"',
    )
    errorif(
        not isinstance(field, FourierCurrentPotentialField),
        ValueError,
        "Expected FourierCurrentPotentialField for field argument, instead got type "
        f"{type(field)}",
    )
    warnif(
        field.sym_Phi == "cos" and field.sym is True and eq.sym is True,
        UserWarning,
        "Detected incompatible Phi symmetry (cos) for symmetric"
        " equilibrium and surface geometry, it is recommended to switch to"
        " sin symmetry for Phi.",
    )

    current_potential_field = field.copy()  # copy field so we can modify freely
    M_coil = current_helicity[0]  # poloidal transits before coil returns to itself
    N_coil = current_helicity[1]  # toroidal transits before coil returns to itself

    # maybe it is an EquilibriaFamily
    errorif(hasattr(eq, "__len__"), ValueError, "Expected a single equilibrium")

    if vacuum:
        # check if vacuum flag should be True or not
        pres = np.max(np.abs(eq.compute("p")["p"]))
        curr = np.max(np.abs(eq.compute("current")["current"]))
        warnif(
            pres > 1e-8,
            UserWarning,
            f"Pressure appears to be non-zero (max {pres} Pa), "
            + "vacuum flag should probably be set to False.",
        )
        warnif(
            curr > 1e-8,
            UserWarning,
            f"Current appears to be non-zero (max {curr} A), "
            + "vacuum flag should probably be set to False.",
        )

    data = {}
    if external_field:  # ensure given field is an instance of _MagneticField
        assert hasattr(external_field, "compute_magnetic_field"), (
            "Expected MagneticField for argument external_field, "
            f"got type {type(external_field)} "
        )
        data["external_field"] = external_field
        data["external_field_grid"] = external_field_grid

    if source_grid is None:
        source_grid = LinearGrid(
            M=max(3 * current_potential_field.M_Phi, 30),
            N=max(3 * current_potential_field.N_Phi, 30),
            NFP=int(eq.NFP),
        )
    if eval_grid is None:
        eval_grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=int(eq.NFP))
    B_eq_surf = eq.compute("|B|", eval_grid)["|B|"]
    # just need it for normalization, so do a simple mean
    normalization_B = jnp.mean(B_eq_surf)

    data["eval_grid"] = eval_grid
    data["source_grid"] = source_grid

    # plasma surface normal vector magnitude on eval grid
    ne_mag = eq.compute(["|e_theta x e_zeta|"], eval_grid)["|e_theta x e_zeta|"]
    # winding surface normal vector magnitude on source grid
    ns_mag = current_potential_field.compute(["|e_theta x e_zeta|"], source_grid)[
        "|e_theta x e_zeta|"
    ]

    # calculate net enclosed poloidal and toroidal currents
    G_tot = -(eq.compute("G", grid=source_grid)["G"][0] / mu_0 * 2 * jnp.pi)

    if external_field:
        G_ext = _G_from_external_field(external_field, eq, external_field_grid)
    else:
        G_ext = 0

    # G needed by surface current is the total G minus the external contribution
    G = G_tot - G_ext
    # calculate I, net toroidal current on winding surface
    if N_coil == 0 and M_coil == 0:  # windowpane coils
        I = G = 0
    elif N_coil == 0:  # modular coils
        I = 0
    elif M_coil == 0:  # only toroidally closed coils, like PF coils
        I = N_coil * G_tot  # give some toroidal current corr. to N_coil
        G = 0  # because I==0
    else:  # helical coils
        I = N_coil * G / M_coil

    # define functions which will be differentiated
    def Bn_from_K(phi_mn, I, G):
        """B from K from REGCOIL eqn 4."""
        params = current_potential_field.params_dict
        params["Phi_mn"] = phi_mn
        params["I"] = I
        params["G"] = G
        Bn, _ = current_potential_field.compute_Bnormal(
            eq.surface, eval_grid=eval_grid, source_grid=source_grid, params=params
        )
        return Bn

    def K(phi_mn, I, G):
        params = current_potential_field.params_dict
        params["Phi_mn"] = phi_mn
        params["I"] = I
        params["G"] = G
        data = current_potential_field.compute("K", grid=source_grid, params=params)
        return data["K"]

    def surfint_Ksqd(phi_mn, I, G):
        K_vec = K(phi_mn, I, G)
        K_mag_sqd = dot(K_vec, K_vec, axis=1)
        integrand = (K_mag_sqd * ns_mag * source_grid.weights).T
        return jnp.sum(integrand).squeeze()

    def surfint_Bnsqd(phi_mn, I, G):
        integrand = ((Bn_from_K(phi_mn, I, G) ** 2) * ne_mag * eval_grid.weights).T
        return jnp.sum(integrand).squeeze()

    if regularization_type == "regcoil":
        # also need these gradients for the RHS if using regcoil regularization
        # set jacobian deriv mode based on the matrix dimensions,  which is the output
        # size (grid nodes, which is eval_grid for Bn and 3*source_grid for K)
        # by the input size (the number of Phi modes)
        deriv_mode = (
            "fwd"
            if eval_grid.num_nodes >= 0.5 * current_potential_field.Phi_basis.num_modes
            else "rev"
        )
        grad_Bn = Derivative(Bn_from_K, mode=deriv_mode).compute(
            current_potential_field.Phi_mn, 0.0, 0.0
        )
        # TODO: likely can make the grad_Ksv one into a compute fxn instead of using
        # JAX to compute dK/dPhimn, but this works for now
        deriv_mode = (
            "fwd"
            if 3 * source_grid.num_nodes
            >= 0.5 * current_potential_field.Phi_basis.num_modes
            else "rev"
        )
        grad_Ksv = Derivative(K).compute(current_potential_field.Phi_mn, 0.0, 0.0)

    timer = Timer()
    # calculate the Jacobian matrix A for  Bn_SV = A*Phi_mn
    timer.start("Jacobian Calculation")
    if regularization_type == "regcoil":
        A1 = Derivative(surfint_Bnsqd, mode="hess").compute(
            current_potential_field.Phi_mn, 0.0, 0.0
        )
        A2 = Derivative(surfint_Ksqd, mode="hess").compute(
            current_potential_field.Phi_mn, 0.0, 0.0
        )

    else:
        # set jacobian deriv mode based on the matrix dimensions
        deriv_mode = (
            "fwd"
            if eval_grid.num_nodes >= 0.5 * current_potential_field.Phi_basis.num_modes
            else "rev"
        )
        A = (
            Derivative(Bn_from_K).compute(current_potential_field.Phi_mn, 0.0, 0.0).T
            * ne_mag
            * eval_grid.weights
        ).T
    timer.stop("Jacobian Calculation")
    if verbose > 1:
        timer.disp("Jacobian Calculation")

    current_potential_field.I = float(I)
    current_potential_field.G = float(G)

    # find the normal field from the secular part of the current potential
    # also mutliply by necessary weights and normal vector magnitude
    Bn_GI = (
        Bn_from_K(jnp.zeros_like(current_potential_field.Phi_mn), I, G)
        * ne_mag
        * eval_grid.weights
    )
    if not vacuum:  # get Bn from plasma contribution
        Bn_plasma = compute_B_plasma(eq, eval_grid, vc_source_grid, normal_only=True)
        Bn_plasma = Bn_plasma * ne_mag * eval_grid.weights

    else:
        Bn_plasma = jnp.zeros_like(Bn_GI)  # from plasma current, currently assume is 0
    # find external field's Bnormal contribution
    if external_field:
        Bn_ext, _ = external_field.compute_Bnormal(
            eq.surface, eval_grid=eval_grid, source_grid=external_field_grid
        )
        Bn_ext = Bn_ext * ne_mag * eval_grid.weights

    else:
        Bn_ext = jnp.zeros_like(Bn_GI)

    rhs = Bn_plasma + Bn_ext + Bn_GI
    if regularization_type == "regcoil":
        rhs_B = -2 * ((grad_Bn.T * rhs).T).sum(axis=0)
        dotted_K_d_K_d_Phimn = dot(
            K(
                jnp.zeros_like(current_potential_field.Phi_mn),
                current_potential_field.I,
                current_potential_field.G,
            )[:, :, jnp.newaxis],
            grad_Ksv,
            axis=1,
        )
        rhs_K = -2 * (dotted_K_d_K_d_Phimn.T * ns_mag * source_grid.weights).T.sum(
            axis=0
        )
    lambda_regularizations = np.atleast_1d(lambda_regularization)

    chi2Bs = []
    chi2Ks = []
    K_mags = []
    phi_mns = []
    Bn_arrs = []
    fields = []

    # calculate the Phi_mn which minimizes
    # (chi^2_B + lambda_regularization*chi^2_K) for each lambda_regularization
    # pre-calculate the SVD
    if regularization_type == "simple":
        u, s, vh = jnp.linalg.svd(A, full_matrices=False)
        s_uT = (u * s).T
        s_uT_b = -s_uT @ rhs
        vht = vh.T

    for lambda_regularization in lambda_regularizations:
        printstring = (
            "Calculating Phi_SV for "
            + f"lambda_regularization = {lambda_regularization:1.5e}"
        )
        if verbose > 0:
            print(
                "#" * len(printstring)
                + "\n"
                + printstring
                + "\n"
                + "#" * len(printstring)
            )

        if regularization_type == "simple":
            # calculate Phi_mn with SVD inverse plus the regularization
            phi_mn_opt = vht @ ((1 / (s**2 + lambda_regularization)) * s_uT_b)
        else:
            # solve linear system
            matrix = A1 + lambda_regularization * A2
            rhs = rhs_B + lambda_regularization * rhs_K
            cho_fac_c_and_lower = cho_factor(matrix)
            phi_mn_opt = cho_solve(cho_fac_c_and_lower, rhs)
            if jnp.any(jnp.isnan(phi_mn_opt)):
                print(
                    "Singular linear system encountered at "
                    f"lambda={lambda_regularization}, retrying with pseudoinverse"
                )
                # failed to solve, likely bc matrix is singular, use lstsq instead
                phi_mn_opt = jnp.linalg.lstsq(matrix, rhs)[0]

        phi_mns.append(phi_mn_opt)

        Bn_SV = Bn_from_K(phi_mn_opt, 0.0, 0.0) * ne_mag * eval_grid.weights
        Bn_tot = Bn_SV + Bn_plasma + Bn_GI + Bn_ext

        chi_B = jnp.sum(Bn_tot * Bn_tot / ne_mag / eval_grid.weights)
        chi2Bs.append(chi_B)

        current_potential_field.Phi_mn = phi_mn_opt
        fields.append(current_potential_field.copy())
        K = current_potential_field.compute(["K"], grid=source_grid)["K"]
        K_mag = jnp.linalg.norm(K, axis=-1)
        chi_K = jnp.sum(K_mag * K_mag * ns_mag * source_grid.weights)
        chi2Ks.append(chi_K)
        K_mags.append(K_mag)
        Bn_print = Bn_tot
        Bn_print_normalized = Bn_tot / normalization_B
        Bn_arrs.append(Bn_tot)
        if verbose > 0:
            units = " (T)"
            printstring = f"chi^2 B = {chi_B:1.5e}"
            print(printstring)
            printstring = f"min Bnormal = {jnp.min(np.abs(Bn_print)):1.5e}"
            printstring += units
            print(printstring)
            printstring = f"Max Bnormal = {jnp.max(jnp.abs(Bn_print)):1.5e}"
            printstring += units
            print(printstring)
            printstring = f"Avg Bnormal = {jnp.mean(jnp.abs(Bn_print)):1.5e}"
            printstring += units
            print(printstring)
            units = " (unitless)"
            printstring = f"min Bnormal = {jnp.min(np.abs(Bn_print_normalized)):1.5e}"
            printstring += units
            print(printstring)
            printstring = f"Max Bnormal = {jnp.max(jnp.abs(Bn_print_normalized)):1.5e}"
            printstring += units
            print(printstring)
            printstring = f"Avg Bnormal = {jnp.mean(jnp.abs(Bn_print_normalized)):1.5e}"
            printstring += units
            print(printstring)
    data["lambda_regularization"] = lambda_regularizations
    data["Phi_mn"] = phi_mns
    data["I"] = I
    data["G"] = G
    data["chi^2_B"] = chi2Bs
    data["chi^2_K"] = chi2Ks
    data["|K|"] = K_mags
    data["Bn_total"] = Bn_arrs

    return fields, data


# TODO: replace contour finding with optimizing Winding surface curves
# once that is implemented
def _find_current_potential_contours(
    surface_current_field,
    num_coils,
    npts=128,
    show_plots=False,
    stell_sym=False,
    net_poloidal_current=None,
    net_toroidal_current=None,
    helicity=None,
    plot_kwargs={},
):
    """Find contours of constant current potential (i.e. coils).

    Surface current K is assumed given by

    K = n x ∇ Φ

    Φ(θ,ζ) = Φₛᵥ(θ,ζ) + Gζ/2π + Iθ/2π

    where:

        - n is the winding surface unit normal.
        - Φ is the current potential function, which is a function of theta and zeta,
          and is given as a secular linear term in theta (I)  and zeta (G) and a double
          Fourier series in theta/zeta.

    Parameters
    ----------
    surface_current_field : FourierCurrentPotentialField
        Surface current potential to find contours of
    num_coils : int
        number of contours desired
    npts : int
        number of points to discretize the current potential with in the
        zeta direction. The discretization in theta will be proportional to this.
        The discretized current potential is then passed to the
        ``skimage.measure.find_contours`` which finds contours using a
        marching squares algorithm.
    show_plots : bool, optional
        whether to plot the contours, useful for debugging or seeing why contour finding
        fails, by default False
    stell_sym : bool, optional
        if the modular coilset has stellarator symmetry, by default False.
        Does nothing for helical coils.
    net_poloidal_current, net_toroidal_current : float, optional
        the net poloidal (toroidal) current flowing on the surface. If None, will
         attempt to infer from the given surface_current_field's attributes.
    helicity : int, optional
        the helicity of the coil currents, should be consistent with the passed-in
        net currents. If None, will use the correct ratio of net poloidal and net
        toroidal currents.
    plot_kwargs : tuple
            figsize to pass to matplotlib figure call, to control size of figure
            if ``show_plots=True``

    Returns
    -------
    contour_theta, contour_zeta: list of 1D arrays
        list of length num_coils containing arrays of the theta
        and zeta values describing each contour found.

    """
    ################################################################
    # find current helicity
    ################################################################
    # we know that I = -(G - G_ext) / (helicity * NFP)
    # if net_toroidal_current is zero, then we have modular coils,
    # and just make helicity zero
    net_poloidal_current = setdefault(
        net_poloidal_current, surface_current_field.G, net_poloidal_current
    )
    net_toroidal_current = setdefault(
        net_toroidal_current, surface_current_field.I, net_toroidal_current
    )
    nfp = surface_current_field.NFP
    helicity = setdefault(
        helicity,
        safediv(net_poloidal_current, net_toroidal_current * nfp, threshold=1e-8),
        helicity,
    )
    coil_type = "modular" if jnp.isclose(helicity, 0) else "helical"
    dz = 2 * np.pi / nfp / npts
    if coil_type == "helical":
        # helical coils
        zeta_full = jnp.linspace(
            0, 2 * jnp.pi / nfp, round(2 * jnp.pi / nfp / dz), endpoint=True
        )
        # ensure we have always have points at least from -2pi, 2pi as depending
        # on sign of I, the contours from Phi = [0, abs(I)] may have their starting
        # points (the theta value at zeta=0) be positive or negative theta values,
        # and we want to ensure we catch the start and end of the contours
        theta0 = jnp.sign(helicity) * 2 * jnp.pi
        theta1 = -jnp.sign(helicity) * (2 * np.pi * int(np.abs(helicity) + 1))
        theta_full = jnp.linspace(
            theta0,
            theta1,
            round(abs(theta1 - theta0) / dz),
            endpoint=True,
        )

        theta_full = jnp.sort(theta_full)
    else:
        # modular coils
        theta_full = jnp.linspace(0, 2 * jnp.pi, round(2 * jnp.pi / dz))
        # we start below 0 for zeta to allow for contours which may go in/out of
        # the zeta=0 plane
        zeta_full = jnp.linspace(
            -jnp.pi / nfp, (2 + 1) * jnp.pi / nfp, round(4 * jnp.pi / dz)
        )

    ################################################################
    # find contours of constant phi
    ################################################################
    # make linspace contours
    if coil_type == "helical":
        # helical coils
        # we start them on zeta=0 plane, so we will find contours
        # going from 0 to I (corresponding to zeta=0, and theta*sign(I) increasing)
        contours = jnp.linspace(
            0, jnp.abs(net_toroidal_current), num_coils + 1, endpoint=True
        )
        contours = jnp.sort(contours)
    else:
        # modular coils
        # go from zero to G/nfp
        # or G/nfp/2 if stell_sym is True
        max_curr = (
            jnp.abs(net_poloidal_current) / nfp / 2
            if stell_sym
            else jnp.abs(net_poloidal_current) / nfp
        )
        # For stell_sym, must pick these carefully
        # so that the coilset has coils that are in order
        # of ascending phi, otherwise we may get the
        # problem that the first coil is over the zeta=0 sym line
        # and the last coil is on the zeta=pi/NFP sym line
        # use a small offset to avoid this issue
        offset = max_curr / num_coils / 2 if stell_sym else 0
        contours = jnp.linspace(
            offset,
            max_curr + offset,
            num_coils + 1,
            endpoint=True,
        ) * jnp.sign(net_poloidal_current)
        contours = jnp.sort(contours)

    theta_full_2D, zeta_full_2D = jnp.meshgrid(theta_full, zeta_full, indexing="ij")

    grid = Grid(
        jnp.vstack(
            (
                jnp.zeros_like(theta_full_2D.flatten(order="F")),
                theta_full_2D.flatten(order="F"),
                zeta_full_2D.flatten(order="F"),
            )
        ).T,
        sort=False,
    )
    phi_total_full = surface_current_field.compute("Phi", grid=grid)["Phi"].reshape(
        theta_full.size, zeta_full.size, order="F"
    )

    # list of arrays of the zeta and theta coordinates
    # of each constant potential contour
    contour_zeta = []
    contour_theta = []
    # list of arrays of the current potential contours,
    # given as indices in theta,zeta
    contours_indices = []
    for contour in contours:
        this_contour = skimage.measure.find_contours(
            np.asarray(jnp.transpose(phi_total_full)), level=contour
        )
        warnif(
            len(this_contour) > 1,
            UserWarning,
            "Detected multiple current potential contours for the same value,"
            + " indicates there may be unclosed coils or window-pane-like structures"
            " for which coil-cutting is not currently supported",
        )
        contours_indices.append(this_contour[0])
    # from the indices, calculate the actual zeta and theta values of the contours
    contours_theta_zeta_not_uniform = [
        np.array(
            [
                np.interp(contour[:, 0], np.arange(zeta_full.size), zeta_full),
                np.interp(contour[:, 1], np.arange(theta_full.size), theta_full),
            ]
        ).T
        for contour in contours_indices
    ]
    # make all the contours the same length by interpolating
    # any shorter ones to match the length of the longest
    Npts = np.max([c.shape[0] for c in contours_indices])
    s = np.linspace(0, 2 * np.pi, Npts)
    contours_theta_zeta = [
        np.array(
            [
                np.interp(s, np.linspace(0, 2 * np.pi, c[:, 0].size), c[:, 0]),
                np.interp(s, np.linspace(0, 2 * np.pi, c[:, 1].size), c[:, 1]),
            ]
        ).T
        for c in contours_theta_zeta_not_uniform
    ]
    # reverse so that our coils start at zeta=0, this is just a choice
    contours_theta_zeta.reverse()

    numCoils = 0
    # to be used to check closure conditions on the coils
    ## closure condition in zeta for modular is returns to same zeta,
    ## while for helical is that the contour dzeta = 2pi/NFP
    zeta_diff = 2 * jnp.pi / nfp if coil_type == "helical" else 0.0
    ## closure condition in theta for modular is that dtheta = 2pi,
    ## while for helical the dtheta = 2pi*abs(helicity)
    theta_diff = (
        2 * jnp.pi * jnp.abs(helicity) if coil_type == "helical" else 2 * jnp.pi
    )

    if show_plots:
        plt.figure(figsize=plot_kwargs.pop("figsize", (8, 6)))
        plt.contourf(
            zeta_full_2D.T,
            theta_full_2D.T,
            jnp.transpose(phi_total_full),
            levels=100,
            **plot_kwargs,
        )
        plt.xlabel(r"$\zeta$")
        plt.ylabel(r"$\theta$")
        plt.xlim([np.min(zeta_full), np.max(zeta_full)])
        plt.ylim([np.min(theta_full), np.max(theta_full)])

    for j in range(num_coils):
        contour_zeta.append(contours_theta_zeta[j][:, 0])
        contour_theta.append(contours_theta_zeta[j][:, 1])
        # check if closed and if not throw warning

        if not jnp.isclose(
            jnp.abs(contour_zeta[-1][-1] - contour_zeta[-1][0]), zeta_diff, rtol=1e-4
        ) or not jnp.isclose(
            jnp.abs(contour_theta[-1][-1] - contour_theta[-1][0]), theta_diff, rtol=1e-4
        ):
            warnings.warn(
                f"Detected a coil contour (coil index {j}) that may not be "
                "closed, this may lead to incorrect coils, "
                "check that the surface current potential contours do not contain "
                "any local maxima or window-pane-like structures,"
                " and that the current potential contours do not go across "
                "The edges of the zeta extent used for the plotting:"
                "the zeta=0 or zeta=2pi/NFP planes for helical coils or the"
                " zeta=-pi/NFP and zeta=2pi+pi/NFP planes, for modular coils. "
                "Use `show_plots=True` to visualize the contours.",
                UserWarning,
            )
            print(f"zeta diff = {jnp.abs(contour_zeta[-1][-1] - contour_zeta[-1][0])}")
            print(f"expected zeta diff = {zeta_diff}")
            print(
                f"theta diff = {jnp.abs(contour_theta[-1][-1] - contour_theta[-1][0])}"
            )
            print(f"expected theta diff = {theta_diff}")

        numCoils += 1
        if show_plots:
            plt.plot(contour_zeta[-1], contour_theta[-1], "-r", linewidth=1)
            if j > 0:
                plt.plot(contour_zeta[-1][-1], contour_theta[-1][-1], "sk")
            else:
                plt.plot(
                    contour_zeta[-1][-1],
                    contour_theta[-1][-1],
                    "sk",
                    label="start of contour",
                )
    if coil_type == "helical":
        # right now these are only over 1 FP
        # so must tile them s.t. they are full coils, by repeating them
        #  with a 2pi/NFP shift in zeta
        # and a -2pi*helicity shift in theta
        # we could alternatively wait until we are in real space and then
        # rotate the coils there, but this also works
        for i_contour in range(num_coils):
            # check if the contour is arranged with zeta=0 at the start
            # or at the end, easiest to do this tiling if we assume
            # the first index is at zeta=0
            zeta_starts_at_zero = (
                contour_zeta[i_contour][-1] > contour_zeta[i_contour][0]
            )
            orig_theta = contour_theta[i_contour]
            orig_zeta = contour_zeta[i_contour]
            if not zeta_starts_at_zero:
                # flip so that the contour starts at zeta=0
                orig_theta = jnp.flip(orig_theta)
                orig_zeta = jnp.flip(orig_zeta)
            orig_endpoint_theta = orig_theta[-1]

            # dont need last points here since we will shift the whole
            # curve over, and we know the last point must be
            # (zeta0+2pi/NFP, theta0+2pi*abs(helicity)),
            # so easiest to just not include them initially and shift whole curve
            orig_theta = jnp.atleast_1d(orig_theta[:-1])
            orig_zeta = jnp.atleast_1d(orig_zeta[:-1])

            contour_theta[i_contour] = jnp.atleast_1d(orig_theta)
            contour_zeta[i_contour] = jnp.atleast_1d(orig_zeta)

            theta_shift = -2 * np.pi * helicity

            zeta_shift = 2 * jnp.pi / nfp - orig_zeta[0]

            for i in range(1, nfp):
                contour_theta[i_contour] = jnp.concatenate(
                    [contour_theta[i_contour], orig_theta + theta_shift * i]
                )
                contour_zeta[i_contour] = jnp.concatenate(
                    [contour_zeta[i_contour], orig_zeta + zeta_shift * i]
                )
            contour_theta[i_contour] = jnp.append(
                contour_theta[i_contour],
                nfp * (orig_endpoint_theta - contour_theta[i_contour][0])
                + contour_theta[i_contour][0],
            )
            contour_zeta[i_contour] = jnp.append(contour_zeta[i_contour], 2 * jnp.pi)
    if show_plots:
        plt.legend()

    return contour_theta, contour_zeta


def _find_XYZ_points(
    theta_pts,
    zeta_pts,
    surface,
):
    contour_X = []
    contour_Y = []
    contour_Z = []

    for thetas, zetas in zip(theta_pts, zeta_pts):
        coords = surface.compute(
            "x",
            grid=Grid(
                jnp.vstack((jnp.zeros_like(thetas), thetas, zetas)).T,
                sort=False,
            ),
            basis="xyz",
        )["x"]
        contour_X.append(coords[:, 0])
        contour_Y.append(coords[:, 1])
        contour_Z.append(coords[:, 2])

    return contour_X, contour_Y, contour_Z


def _G_from_external_field(external_field, eq, external_field_grid):
    # calculate the portion of G provided by external field
    # by integrating external toroidal field along a curve of constant theta
    try:
        G_ext = external_field.G
    except AttributeError:
        curve_grid = LinearGrid(
            N=int(eq.NFP) * 50,
            theta=jnp.array(jnp.pi),  # does not matter which theta we choose
            rho=jnp.array(1.0),
            endpoint=True,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            # ignore warning from unequal NFP for grid and basis,
            # as we don't know a-priori if the external field
            # shares the same discrete symmetry as the equilibrium,
            # so we will use a grid with NFP=1 to be safe
            curve_data = eq.compute(
                ["R", "phi", "Z", "e_zeta"],
                grid=curve_grid,
            )
            curve_coords = jnp.vstack(
                (curve_data["R"], curve_data["phi"], curve_data["Z"])
            ).T
            ext_field_along_curve = external_field.compute_magnetic_field(
                curve_coords, basis="rpz", source_grid=external_field_grid
            )
        # calculate covariant B_zeta = B dot e_zeta from external field
        ext_field_B_zeta = dot(ext_field_along_curve, curve_data["e_zeta"], axis=-1)

        # negative sign here because with REGCOIL convention, negative G makes
        # positive toroidal B
        G_ext = -jnp.sum(ext_field_B_zeta) * 2 * jnp.pi / curve_grid.num_nodes / mu_0
    return G_ext



"""Compute functions for omnigenity objectives.

Notes
-----
Some quantities require additional work to compute at the magnetic axis.
A Python lambda function is used to lazily compute the magnetic axis limits
of these quantities. These lambda functions are evaluated only when the
computational grid has a node on the magnetic axis to avoid potentially
expensive computations.
"""

from interpax import interp1d

from desc.backend import jnp, sign, vmap

from ..utils import cross, dot, safediv
from .data_index import register_compute_fun


@register_compute_fun(
    name="B_theta_mn",
    label="B_{\\theta, m, n}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Fourier coefficients for covariant poloidal component of "
    "magnetic field.",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["B_theta"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _B_theta_mn(params, transforms, profiles, data, **kwargs):
    B_theta = transforms["grid"].meshgrid_reshape(data["B_theta"], "rtz")

    def fitfun(x):
        return transforms["B"].fit(x.flatten(order="F"))

    B_theta_mn = vmap(fitfun)(B_theta)
    # modes stored as shape(rho, mn) flattened
    data["B_theta_mn"] = B_theta_mn.flatten()
    return data


# TODO (#568): do math to change definition of nu so that we can just use B_zeta_mn here
@register_compute_fun(
    name="B_phi_mn",
    label="B_{\\phi, m, n}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Fourier coefficients for covariant toroidal component of "
    "magnetic field in (ρ,θ,ϕ) coordinates.",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["B_phi|r,t"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    aliases="B_zeta_mn",  # TODO(#568): remove when phi != zeta
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _B_phi_mn(params, transforms, profiles, data, **kwargs):
    B_phi = transforms["grid"].meshgrid_reshape(data["B_phi|r,t"], "rtz")

    def fitfun(x):
        return transforms["B"].fit(x.flatten(order="F"))

    B_zeta_mn = vmap(fitfun)(B_phi)
    # modes stored as shape(rho, mn) flattened
    data["B_phi_mn"] = B_zeta_mn.flatten()
    return data


@register_compute_fun(
    name="w_Boozer_mn",
    label="w_{\\mathrm{Boozer},m,n}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="RHS of eq 10 in Hirshman 1995 'Transformation from VMEC to "
    + "Boozer Coordinates'",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 0]], "B": [[0, 0, 0]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["B_theta_mn", "B_phi_mn"],
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _w_mn(params, transforms, profiles, data, **kwargs):
    w_mn = jnp.zeros((transforms["grid"].num_rho, transforms["w"].basis.num_modes))
    Bm = transforms["B"].basis.modes[:, 1]
    Bn = transforms["B"].basis.modes[:, 2]
    wm = transforms["w"].basis.modes[:, 1]
    wn = transforms["w"].basis.modes[:, 2]
    NFP = transforms["w"].basis.NFP
    mask_t = (Bm[:, None] == -wm) & (Bn[:, None] == wn) & (wm != 0)
    mask_z = (Bm[:, None] == wm) & (Bn[:, None] == -wn) & (wm == 0) & (wn != 0)

    num_t = (mask_t @ sign(wn)) * data["B_theta_mn"].reshape(
        (transforms["grid"].num_rho, -1)
    )
    den_t = mask_t @ jnp.abs(wm)
    num_z = (mask_z @ sign(wm)) * data["B_phi_mn"].reshape(
        (transforms["grid"].num_rho, -1)
    )
    den_z = mask_z @ jnp.abs(NFP * wn)

    w_mn = jnp.where(mask_t.any(axis=0), (mask_t.T @ safediv(num_t, den_t).T).T, w_mn)
    w_mn = jnp.where(mask_z.any(axis=0), (mask_z.T @ safediv(num_z, den_z).T).T, w_mn)

    data["w_Boozer_mn"] = w_mn.flatten()
    return data


@register_compute_fun(
    name="w_Boozer",
    label="w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates'",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 0]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _w(params, transforms, profiles, data, **kwargs):
    grid = transforms["grid"]
    w_mn = data["w_Boozer_mn"].reshape((grid.num_rho, -1))
    w = vmap(transforms["w"].transform)(w_mn)  # shape(rho, theta*zeta)
    w = w.reshape((grid.num_rho, grid.num_theta, grid.num_zeta), order="F")
    w = jnp.moveaxis(w, 0, 1)
    data["w_Boozer"] = w.flatten(order="F")
    return data


@register_compute_fun(
    name="w_Boozer_t",
    label="\\partial_{\\theta} w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates', poloidal derivative",
    dim=1,
    params=[],
    transforms={"w": [[0, 1, 0]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _w_t(params, transforms, profiles, data, **kwargs):
    grid = transforms["grid"]
    w_mn = data["w_Boozer_mn"].reshape((grid.num_rho, -1))
    # need to close over dt which can't be vmapped
    fun = lambda x: transforms["w"].transform(x, dt=1)
    w_t = vmap(fun)(w_mn)  # shape(rho, theta*zeta)
    w_t = w_t.reshape((grid.num_rho, grid.num_theta, grid.num_zeta), order="F")
    w_t = jnp.moveaxis(w_t, 0, 1)
    data["w_Boozer_t"] = w_t.flatten(order="F")
    return data


@register_compute_fun(
    name="w_Boozer_z",
    label="\\partial_{\\zeta} w_{\\mathrm{Boozer}}",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Inverse Fourier transform of RHS of eq 10 in Hirshman 1995 "
    + "'Transformation from VMEC to Boozer Coordinates', toroidal derivative",
    dim=1,
    params=[],
    transforms={"w": [[0, 0, 1]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_mn"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _w_z(params, transforms, profiles, data, **kwargs):
    grid = transforms["grid"]
    w_mn = data["w_Boozer_mn"].reshape((grid.num_rho, -1))
    # need to close over dz which can't be vmapped
    fun = lambda x: transforms["w"].transform(x, dz=1)
    w_z = vmap(fun)(w_mn)  # shape(rho, theta*zeta)
    w_z = w_z.reshape((grid.num_rho, grid.num_theta, grid.num_zeta), order="F")
    w_z = jnp.moveaxis(w_z, 0, 1)
    data["w_Boozer_z"] = w_z.flatten(order="F")
    return data


@register_compute_fun(
    name="nu",
    label="\\nu = \\zeta_{B} - \\zeta",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer", "G", "I", "iota", "lambda"],
)
def _nu(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu"] = (data["w_Boozer"] - data["I"] * data["lambda"]) / GI
    return data


@register_compute_fun(
    name="nu_t",
    label="\\partial_{\\theta} \\nu",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function, derivative wrt poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_t", "G", "I", "iota", "lambda_t"],
)
def _nu_t(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu_t"] = (data["w_Boozer_t"] - data["I"] * data["lambda_t"]) / GI
    return data


@register_compute_fun(
    name="nu_z",
    label="\\partial_{\\zeta} \\nu",
    units="rad",
    units_long="radians",
    description="Boozer toroidal stream function, derivative wrt toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["w_Boozer_z", "G", "I", "iota", "lambda_z"],
)
def _nu_z(params, transforms, profiles, data, **kwargs):
    GI = data["G"] + data["iota"] * data["I"]
    data["nu_z"] = (data["w_Boozer_z"] - data["I"] * data["lambda_z"]) / GI
    return data


@register_compute_fun(
    name="theta_B",
    label="\\theta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer poloidal angular coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST", "iota", "nu"],
)
def _theta_B(params, transforms, profiles, data, **kwargs):
    data["theta_B"] = data["theta_PEST"] + data["iota"] * data["nu"]
    return data


@register_compute_fun(
    name="zeta_B",
    label="\\zeta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer toroidal angular coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["phi", "nu"],
)
def _zeta_B(params, transforms, profiles, data, **kwargs):
    data["zeta_B"] = data["phi"] + data["nu"]
    return data


@register_compute_fun(
    name="sqrt(g)_B",
    label="\\sqrt{g}_{B}",
    units="~",
    units_long="None",
    description="Jacobian determinant from Boozer to DESC coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_t", "theta_PEST_z", "phi_t", "phi_z", "nu_t", "nu_z", "iota"],
)
def _sqrtg_B(params, transforms, profiles, data, **kwargs):
    data["sqrt(g)_B"] = (
        data["theta_PEST_t"] * (data["phi_z"] + data["nu_z"])
        - data["theta_PEST_z"] * (data["phi_t"] + data["nu_t"])
        + data["iota"] * (data["nu_t"] * data["phi_z"] - data["nu_z"] * data["phi_t"])
    )
    return data


@register_compute_fun(
    name="|B|_mn",
    label="B_{mn}^{\\mathrm{Boozer}}",
    units="T",
    units_long="Tesla",
    description="Boozer harmonics of magnetic field",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]], "grid": []},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)_B", "|B|", "rho", "theta_B", "zeta_B"],
    resolution_requirement="tz",
    grid_requirement={"is_meshgrid": True},
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _B_mn(params, transforms, profiles, data, **kwargs):
    norm = 2 ** (3 - jnp.sum((transforms["B"].basis.modes == 0), axis=1))
    grid = transforms["grid"]

    def fun(rho, theta_B, zeta_B, sqrtg_B, B):
        # this fits Boozer modes on a single surface
        nodes = jnp.array([rho, theta_B, zeta_B]).T
        B_mn = (
            norm  # 1 if m=n=0, 2 if m=0 or n=0, 4 if m!=0 and n!=0
            * (transforms["B"].basis.evaluate(nodes).T @ (sqrtg_B * B))
            / transforms["B"].grid.num_nodes
        )
        return B_mn

    def reshape(x):
        return grid.meshgrid_reshape(x, "rtz").reshape((grid.num_rho, -1))

    rho, theta_B, zeta_B, sqrtg_B, B = map(
        reshape,
        (data["rho"], data["theta_B"], data["zeta_B"], data["sqrt(g)_B"], data["|B|"]),
    )
    B_mn = vmap(fun)(rho, theta_B, zeta_B, sqrtg_B, B)
    data["|B|_mn"] = B_mn.flatten()
    return data


@register_compute_fun(
    name="B modes",
    label="\\mathrm{Boozer~modes}",
    units="~",
    units_long="None",
    description="Boozer harmonics",
    dim=1,
    params=[],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    M_booz="int: Maximum poloidal mode number for Boozer harmonics. Default 2*eq.M",
    N_booz="int: Maximum toroidal mode number for Boozer harmonics. Default 2*eq.N",
)
def _B_modes(params, transforms, profiles, data, **kwargs):
    data["B modes"] = transforms["B"].basis.modes
    return data


@register_compute_fun(
    name="f_C",
    label="[(M \\iota - N) (\\mathbf{B} \\times \\nabla \\psi)"
    + " - (M G + N I) \\mathbf{B}] \\cdot \\nabla B",
    units="T^{3}",
    units_long="Tesla cubed",
    description="Two-term quasisymmetry metric",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "iota",
        "psi_r/sqrt(g)",
        "B_theta",
        "B_zeta",
        "|B|_t",
        "|B|_z",
        "G",
        "I",
        "B*grad(|B|)",
    ],
    helicity="tuple: Type of quasisymmetry, (M,N). Default (1,0)",
)
def _f_C(params, transforms, profiles, data, **kwargs):
    M, N = kwargs.get("helicity", (1, 0))
    data["f_C"] = (M * data["iota"] - N) * data["psi_r/sqrt(g)"] * (
        data["B_zeta"] * data["|B|_t"] - data["B_theta"] * data["|B|_z"]
    ) - (M * data["G"] + N * data["I"]) * data["B*grad(|B|)"]
    return data


@register_compute_fun(
    name="f_T",
    label="\\nabla \\psi \\times \\nabla B \\cdot \\nabla "
    + "(\\mathbf{B} \\cdot \\nabla B)",
    units="T^{4} \\cdot m^{-2}",
    units_long="Tesla quarted / square meters",
    description="Triple product quasisymmetry metric",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["psi_r/sqrt(g)", "|B|_t", "|B|_z", "(B*grad(|B|))_t", "(B*grad(|B|))_z"],
)
def _f_T(params, transforms, profiles, data, **kwargs):
    data["f_T"] = data["psi_r/sqrt(g)"] * (
        data["|B|_t"] * data["(B*grad(|B|))_z"]
        - data["|B|_z"] * data["(B*grad(|B|))_t"]
    )
    return data


@register_compute_fun(
    name="eta",
    label="\\eta",
    units="rad",
    units_long="radians",
    description="Intermediate omnigenity coordinate along field lines",
    dim=1,
    params=[],
    transforms={"h": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization="desc.magnetic_fields._core.OmnigenousField",
)
def _eta(params, transforms, profiles, data, **kwargs):
    data["eta"] = transforms["h"].grid.nodes[:, 1]
    return data


@register_compute_fun(
    name="alpha",
    label="\\alpha",
    units="rad",
    units_long="radians",
    description="Field line label, defined on [0, 2pi)",
    dim=1,
    params=[],
    transforms={"h": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization="desc.magnetic_fields._core.OmnigenousField",
)
def _alpha(params, transforms, profiles, data, **kwargs):
    data["alpha"] = transforms["h"].grid.nodes[:, 2]
    return data


@register_compute_fun(
    name="h",
    label="h = \\theta + (N / M) \\zeta",
    units="rad",
    units_long="radians",
    description="Omnigenity symmetry angle",
    dim=1,
    params=["x_lmn"],
    transforms={"h": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["eta"],
    resolution_requirement="tz",
    parameterization="desc.magnetic_fields._core.OmnigenousField",
)
def _omni_angle(params, transforms, profiles, data, **kwargs):
    data["h"] = transforms["h"].transform(params["x_lmn"]) + 2 * data["eta"] + jnp.pi
    return data


@register_compute_fun(
    name="theta_B",
    label="\\theta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer poloidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["alpha", "h"],
    parameterization="desc.magnetic_fields._core.OmnigenousField",
    helicity="tuple: Type of quasisymmetry, (M,N). Default (1,0)",
    iota="float: Value of rotational transform on the Omnigenous surface. Default 1.0",
)
def _omni_map_theta_B(params, transforms, profiles, data, **kwargs):
    M, N = kwargs.get("helicity", (1, 0))
    iota = kwargs.get("iota", 1)

    # coordinate mapping matrix from (alpha,h) to (theta_B,zeta_B)
    # need a bunch of wheres to avoid division by zero causing NaN in backward pass
    # this is fine since the incorrect values get ignored later, except in OT or OH
    # where fieldlines are exactly parallel to |B| contours, but this is a degenerate
    # case of measure 0 so this kludge shouldn't affect things too much.
    mat_OP = jnp.array(
        [[N, iota / jnp.where(N == 0, 1, N)], [0, 1 / jnp.where(N == 0, 1, N)]]
    )
    mat_OT = jnp.array([[0, -1], [M, -1 / jnp.where(iota == 0, 1.0, iota)]])
    den = jnp.where((N - M * iota) == 0, 1.0, (N - M * iota))
    mat_OH = jnp.array([[N, M * iota / den], [M, M / den]])
    matrix = jnp.where(
        M == 0,
        mat_OP,
        jnp.where(
            N == 0,
            mat_OT,
            mat_OH,
        ),
    )

    # solve for (theta_B,zeta_B) corresponding to (eta,alpha)
    booz = matrix @ jnp.vstack((data["alpha"], data["h"]))
    data["theta_B"] = booz[0, :]
    data["zeta_B"] = booz[1, :]
    return data


@register_compute_fun(
    name="zeta_B",
    label="\\zeta_{B}",
    units="rad",
    units_long="radians",
    description="Boozer toroidal angle",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_B"],
    parameterization="desc.magnetic_fields._core.OmnigenousField",
)
def _omni_map_zeta_B(params, transforms, profiles, data, **kwargs):
    return data  # noqa: unused dependency


@register_compute_fun(
    name="|B|",
    label="|\\mathbf{B}|",
    units="T",
    units_long="Tesla",
    description="Magnitude of omnigenous magnetic field",
    dim=1,
    params=["B_lm"],
    transforms={"B": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=["eta"],
    parameterization="desc.magnetic_fields._core.OmnigenousField",
)
def _B_omni(params, transforms, profiles, data, **kwargs):
    # reshaped to size (L_B, M_B)
    B_lm = params["B_lm"].reshape((transforms["B"].basis.L + 1, -1))
    # assuming single flux surface, so only take first row (single node)
    B_input = vmap(lambda x: transforms["B"].transform(x))(B_lm.T)[:, 0]
    B_input = jnp.sort(B_input)  # sort to ensure monotonicity
    eta_input = jnp.linspace(0, jnp.pi / 2, num=B_input.size)

    # |B|_omnigeneous is an even function so B(-eta) = B(+eta) = B(|eta|)
    data["|B|"] = interp1d(
        jnp.abs(data["eta"]), eta_input, B_input, method="monotonic-0"
    )
    return data


@register_compute_fun(
    name="isodynamicity",
    label="1/|B|^2 (\\mathbf{b} \\times \\nabla B) \\cdot \\nabla \\psi",
    units="~",
    units_long="None",
    description="Measure of cross field drift at each point, "
    + "unweighted by particle energy",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["b", "grad(|B|)", "|B|", "grad(psi)"],
)
def _isodynamicity(params, transforms, profiles, data, **kwargs):
    data["isodynamicity"] = (
        dot(cross(data["b"], data["grad(|B|)"]), data["grad(psi)"]) / data["|B|"] ** 2
    )
    return data



"""Methods for computing bounce integrals (singular or otherwise)."""

from interpax import CubicHermiteSpline, PPoly
from orthax.legendre import leggauss

from desc.backend import jnp
from desc.integrals.bounce_utils import (
    _bounce_quadrature,
    _check_bounce_points,
    _set_default_plot_kwargs,
    bounce_points,
    get_pitch_inv_quad,
    interp_to_argmin,
    plot_ppoly,
)
from desc.integrals.interp_utils import polyder_vec
from desc.integrals.quad_utils import (
    automorphism_sin,
    get_quadrature,
    grad_automorphism_sin,
)
from desc.io import IOAble
from desc.utils import errorif, setdefault, warnif


class Bounce1D(IOAble):
    """Computes bounce integrals using one-dimensional local spline methods.

    The bounce integral is defined as ∫ f(λ, ℓ) dℓ, where
        dℓ parameterizes the distance along the field line in meters,
        f(λ, ℓ) is the quantity to integrate along the field line,
        and the boundaries of the integral are bounce points ℓ₁, ℓ₂ s.t. λ|B|(ℓᵢ) = 1,
        where λ is a constant defining the integral proportional to the magnetic moment
        over energy and |B| is the norm of the magnetic field.

    For a particle with fixed λ, bounce points are defined to be the location on the
    field line such that the particle's velocity parallel to the magnetic field is zero.
    The bounce integral is defined up to a sign. We choose the sign that corresponds to
    the particle's guiding center trajectory traveling in the direction of increasing
    field-line-following coordinate ζ.

    Notes
    -----
    Brief description of algorithm for developers.

    For applications which reduce to computing a nonlinear function of distance
    along field lines between bounce points, it is required to identify these
    points with field-line-following coordinates. (In the special case of a linear
    function summing integrals between bounce points over a flux surface, arbitrary
    coordinate systems may be used as that task reduces to a surface integral,
    which is invariant to the order of summation).

    The DESC coordinate system is related to field-line-following coordinate
    systems by a relation whose solution is best found with Newton iteration.
    There is a unique real solution to that relation, so Newton iteration is a
    globally convergent root-finding algorithm here. For the task of finding
    bounce points, Newton iteration is not a globally convergent algorithm to
    find the real roots of r : ζ ↦ |B|(ζ) − 1/λ where ζ is a field-line-following
    coordinate. For this, function approximation of |B| is necessary.

    The function approximation in ``Bounce1D`` is ignorant that the objects to
    approximate are defined on a bounded subset of ℝ². Instead, the domain is
    projected to ℝ, where information sampled about the function at infinity
    cannot support reconstruction of the function near the origin. As the
    functions of interest do not vanish at infinity, pseudo-spectral techniques
    are not used. Instead, function approximation is done with local splines.
    This is useful if one can efficiently obtain data along field lines and the
    number of toroidal transits to follow a field line is not large.

    After computing the bounce points, the supplied quadrature is performed.
    By default, this is a Gauss quadrature after removing the singularity.
    Local splines interpolate functions in the integrand to the quadrature nodes.

    See Also
    --------
    Bounce2D
        Uses two-dimensional pseudo-spectral techniques for the same task.

    Examples
    --------
    See ``tests/test_integrals.py::TestBounce1D::test_bounce1d_checks``.

    Attributes
    ----------
    required_names : list
        Names in ``data_index`` required to compute bounce integrals.
    B : jnp.ndarray
        Shape (M, L, N - 1, B.shape[-1]).
        Polynomial coefficients of the spline of |B| in local power basis.
        Last axis enumerates the coefficients of power series. For a polynomial
        given by ∑ᵢⁿ cᵢ xⁱ, coefficient cᵢ is stored at ``B[...,n-i]``.
        Third axis enumerates the polynomials that compose a particular spline.
        Second axis enumerates flux surfaces.
        First axis enumerates field lines of a particular flux surface.

    """

    required_names = ["B^zeta", "B^zeta_z|r,a", "|B|", "|B|_z|r,a"]
    get_pitch_inv_quad = staticmethod(get_pitch_inv_quad)

    def __init__(
        self,
        grid,
        data,
        quad=leggauss(32),
        automorphism=(automorphism_sin, grad_automorphism_sin),
        *,
        Bref=1.0,
        Lref=1.0,
        is_reshaped=False,
        check=False,
        **kwargs,
    ):
        """Returns an object to compute bounce integrals.

        Parameters
        ----------
        grid : Grid
            Tensor-product grid in (ρ, α, ζ) Clebsch coordinates.
            The ζ coordinates (the unique values prior to taking the tensor-product)
            must be strictly increasing and preferably uniformly spaced. These are used
            as knots to construct splines. A reference knot density is 100 knots per
            toroidal transit. Note that below shape notation defines
            L = ``grid.num_rho``, M = ``grid.num_alpha``, and N = ``grid.num_zeta``.
        data : dict[str, jnp.ndarray]
            Data evaluated on ``grid``.
            Must include names in ``Bounce1D.required_names``.
        quad : tuple[jnp.ndarray]
            Quadrature points xₖ and weights wₖ for the approximate evaluation of an
            integral ∫₋₁¹ g(x) dx = ∑ₖ wₖ g(xₖ). Default is 32 points.
            For weak singular integrals, use ``chebgauss2`` from
            ``desc.integrals.quad_utils``.
            For strong singular integrals, use ``leggauss``.
        automorphism : tuple[Callable] or None
            The first callable should be an automorphism of the real interval [-1, 1].
            The second callable should be the derivative of the first. This map defines
            a change of variable for the bounce integral. The choice made for the
            automorphism will affect the performance of the quadrature method.
            For weak singular integrals, use ``None``.
            For strong singular integrals, use
            ``(automorphism_sin,grad_automorphism_sin)`` from
        Bref : float
            Optional. Reference magnetic field strength for normalization.
        Lref : float
            Optional. Reference length scale for normalization.
        is_reshaped : bool
            Whether the arrays in ``data`` are already reshaped to the expected form of
            shape (..., N) or (..., L, N) or (M, L, N). This option can be used to
            iteratively compute bounce integrals one field line or one flux surface
            at a time, respectively, reducing memory usage. To do so, set to true and
            provide only those axes of the reshaped data. Default is false.
        check : bool
            Flag for debugging. Must be false for JAX transformations.

        """
        # Strictly increasing zeta knots enforces dζ > 0.
        # To retain dℓ = (|B|/B^ζ) dζ > 0 after fixing dζ > 0, we require
        # B^ζ = B⋅∇ζ > 0. This is equivalent to changing the sign of ∇ζ or [∂ℓ/∂ζ]|ρ,a.
        # Recall dζ = ∇ζ⋅dR, implying 1 = ∇ζ⋅(e_ζ|ρ,a). Hence, a sign change in ∇ζ
        # requires the same sign change in e_ζ|ρ,a to retain the metric identity.
        warnif(
            check and kwargs.pop("warn", True) and jnp.any(data["B^zeta"] <= 0),
            msg="(∂ℓ/∂ζ)|ρ,a > 0 is required. Enforcing positive B^ζ.",
        )
        data = {
            "B^zeta": jnp.abs(data["B^zeta"]) * Lref / Bref,
            "B^zeta_z|r,a": data["B^zeta_z|r,a"]
            * jnp.sign(data["B^zeta"])
            * Lref
            / Bref,
            "|B|": data["|B|"] / Bref,
            "|B|_z|r,a": data["|B|_z|r,a"] / Bref,  # This is already the correct sign.
        }
        self._data = (
            data
            if is_reshaped
            else dict(zip(data.keys(), Bounce1D.reshape_data(grid, *data.values())))
        )
        self._x, self._w = get_quadrature(quad, automorphism)

        # Compute local splines.
        self._zeta = grid.compress(grid.nodes[:, 2], surface_label="zeta")
        self.B = jnp.moveaxis(
            CubicHermiteSpline(
                x=self._zeta,
                y=self._data["|B|"],
                dydx=self._data["|B|_z|r,a"],
                axis=-1,
                check=check,
            ).c,
            source=(0, 1),
            destination=(-1, -2),
        )
        self._dB_dz = polyder_vec(self.B)

        # Add axis here instead of in ``_bounce_quadrature``.
        for name in self._data:
            self._data[name] = self._data[name][..., jnp.newaxis, :]

    @staticmethod
    def reshape_data(grid, *arys):
        """Reshape arrays for acceptable input to ``integrate``.

        Parameters
        ----------
        grid : Grid
            Tensor-product grid in (ρ, α, ζ) Clebsch coordinates.
        arys : jnp.ndarray
            Data evaluated on grid.

        Returns
        -------
        f : jnp.ndarray
            Shape (M, L, N).
            Reshaped data which may be given to ``integrate``.

        """
        f = [grid.meshgrid_reshape(d, "arz") for d in arys]
        return f if len(f) > 1 else f[0]

    def points(self, pitch_inv, *, num_well=None):
        """Compute bounce points.

        Parameters
        ----------
        pitch_inv : jnp.ndarray
            Shape (M, L, num_pitch).
            1/λ values to compute the bounce points at each field line. 1/λ(α,ρ) is
            specified by ``pitch_inv[α,ρ]`` where in the latter the labels
            are interpreted as the indices that correspond to that field line.
        num_well : int or None
            Specify to return the first ``num_well`` pairs of bounce points for each
            pitch along each field line. This is useful if ``num_well`` tightly
            bounds the actual number. As a reference, there are typically 20 wells
            per toroidal transit for a given pitch. You can check this by plotting
            the field lines with the ``check_points`` method.

            If not specified, then all bounce points are returned. If there were fewer
            wells detected along a field line than the size of the last axis of the
            returned arrays, then that axis is padded with zero.

        Returns
        -------
        z1, z2 : tuple[jnp.ndarray]
            Shape (M, L, num_pitch, num_well).
            Tuple of length two (z1, z2) that stores ζ coordinates of bounce points.
            The points are ordered and grouped such that the straight line path
            between ``z1`` and ``z2`` resides in the epigraph of |B|.

            If there were less than ``num_well`` wells detected along a field line,
            then the last axis, which enumerates bounce points for a particular field
            line and pitch, is padded with zero.

        """
        return bounce_points(pitch_inv, self._zeta, self.B, self._dB_dz, num_well)

    def check_points(self, points, pitch_inv, *, plot=True, **kwargs):
        """Check that bounce points are computed correctly.

        Parameters
        ----------
        points : tuple[jnp.ndarray]
            Shape (M, L, num_pitch, num_well).
            Output of method ``self.points``.
            Tuple of length two (z1, z2) that stores ζ coordinates of bounce points.
            The points are ordered and grouped such that the straight line path
            between ``z1`` and ``z2`` resides in the epigraph of |B|.
        pitch_inv : jnp.ndarray
            Shape (M, L, num_pitch).
            1/λ values to compute the bounce points at each field line. 1/λ(α,ρ) is
            specified by ``pitch_inv[α,ρ]`` where in the latter the labels
            are interpreted as the indices that correspond to that field line.
        plot : bool
            Whether to plot the field lines and bounce points of the given pitch angles.
        kwargs
            Keyword arguments into ``desc/integrals/bounce_utils.py::plot_ppoly``.

        Returns
        -------
        plots : list
            Matplotlib (fig, ax) tuples for the 1D plot of each field line.

        """
        return _check_bounce_points(
            z1=points[0],
            z2=points[1],
            pitch_inv=pitch_inv,
            knots=self._zeta,
            B=self.B,
            plot=plot,
            **kwargs,
        )

    def integrate(
        self,
        integrand,
        pitch_inv,
        f=None,
        weight=None,
        points=None,
        *,
        method="cubic",
        batch=True,
        check=False,
        plot=False,
    ):
        """Bounce integrate ∫ f(λ, ℓ) dℓ.

        Computes the bounce integral ∫ f(λ, ℓ) dℓ for every field line and pitch.

        Parameters
        ----------
        integrand : callable
            The composition operator on the set of functions in ``f`` that maps the
            functions in ``f`` to the integrand f(λ, ℓ) in ∫ f(λ, ℓ) dℓ. It should
            accept the arrays in ``f`` as arguments as well as the additional keyword
            arguments: ``B`` and ``pitch``. A quadrature will be performed to
            approximate the bounce integral of ``integrand(*f,B=B,pitch=pitch)``.
        pitch_inv : jnp.ndarray
            Shape (M, L, num_pitch).
            1/λ values to compute the bounce integrals. 1/λ(α,ρ) is specified by
            ``pitch_inv[α,ρ]`` where in the latter the labels are interpreted
            as the indices that correspond to that field line.
        f : list[jnp.ndarray] or jnp.ndarray
            Shape (M, L, N).
            Real scalar-valued functions evaluated on the ``grid`` supplied to
            construct this object. These functions should be arguments to the callable
            ``integrand``. Use the method ``Bounce1D.reshape_data`` to reshape the data
            into the expected shape.
        weight : jnp.ndarray
            Shape (M, L, N).
            If supplied, the bounce integral labeled by well j is weighted such that
            the returned value is w(j) ∫ f(λ, ℓ) dℓ, where w(j) is ``weight``
            interpolated to the deepest point in that magnetic well. Use the method
            ``Bounce1D.reshape_data`` to reshape the data into the expected shape.
        points : tuple[jnp.ndarray]
            Shape (M, L, num_pitch, num_well).
            Optional, output of method ``self.points``.
            Tuple of length two (z1, z2) that stores ζ coordinates of bounce points.
            The points are ordered and grouped such that the straight line path
            between ``z1`` and ``z2`` resides in the epigraph of |B|.
        method : str
            Method of interpolation.
            See https://interpax.readthedocs.io/en/latest/_api/interpax.interp1d.html.
            Default is cubic C1 local spline.
        batch : bool
            Whether to perform computation in a batched manner. Default is true.
        check : bool
            Flag for debugging. Must be false for JAX transformations.
        plot : bool
            Whether to plot the quantities in the integrand interpolated to the
            quadrature points of each integral. Ignored if ``check`` is false.

        Returns
        -------
        result : jnp.ndarray
            Shape is same as input points.
            Last axis enumerates the bounce integrals for a given field line,
            flux surface, and pitch value.

        """
        if points is None:
            points = self.points(pitch_inv)
        result = _bounce_quadrature(
            x=self._x,
            w=self._w,
            integrand=integrand,
            points=points,
            pitch_inv=pitch_inv,
            f=setdefault(f, []),
            data=self._data,
            knots=self._zeta,
            method=method,
            batch=batch,
            check=check,
            plot=plot,
        )
        if weight is not None:
            result *= interp_to_argmin(
                weight,
                points,
                self._zeta,
                self.B,
                self._dB_dz,
                method,
            )
        return result

    def plot(self, m, l, pitch_inv=None, **kwargs):
        """Plot the field line and bounce points of the given pitch angles.

        Parameters
        ----------
        m, l : int, int
            Indices into the nodes of the grid supplied to make this object.
            ``alpha,rho=grid.meshgrid_reshape(grid.nodes[:,:2],"arz")[m,l,0]``.
        pitch_inv : jnp.ndarray
            Shape (num_pitch, ).
            Optional, 1/λ values whose corresponding bounce points on the field line
            specified by Clebsch coordinate α(m), ρ(l) will be plotted.
        kwargs
            Keyword arguments into ``desc/integrals/bounce_utils.py::plot_ppoly``.

        Returns
        -------
        fig, ax
            Matplotlib (fig, ax) tuple.

        """
        B, dB_dz = self.B, self._dB_dz
        if B.ndim == 4:
            B = B[m]
            dB_dz = dB_dz[m]
        if B.ndim == 3:
            B = B[l]
            dB_dz = dB_dz[l]
        if pitch_inv is not None:
            errorif(
                pitch_inv.ndim > 1,
                msg=f"Got pitch_inv.ndim={pitch_inv.ndim}, but expected 1.",
            )
            z1, z2 = bounce_points(pitch_inv, self._zeta, B, dB_dz)
            kwargs["z1"] = z1
            kwargs["z2"] = z2
            kwargs["k"] = pitch_inv
        fig, ax = plot_ppoly(PPoly(B.T, self._zeta), **_set_default_plot_kwargs(kwargs))
        return fig, ax



"""Classes for representing geometric objects like curves and surfaces."""

from .core import Curve, Surface
from .curve import FourierPlanarCurve, FourierRZCurve, FourierXYZCurve, SplineXYZCurve
from .surface import FourierRZToroidalSurface, ZernikeRZToroidalSection
